import socket 

def get_local_ip():
    """
    heuristic for determining an ip address to connect to the grpc service
    :return: ip address in string form
    """
    with socket.socket(socket.AF_INET, socket.SOCK_DGRAM) as s:
        try:
            # doesn't even have to be reachable
            s.connect(('10.255.255.255', 1))
        except socket.error:
            return '127.0.0.1'

        return s.getsockname()[0]

ip=get_local_ip()
port = 50052
environment = "windows"
configuration ="release"



import os
import socket

from ansys.dpf.core._version import __version__

# environment variables for pyansys.com
if 'jupyter' in socket.gethostname():
    if 'ANSYS_PATH' not in os.environ:
        os.environ['ANSYS_PATH'] = '/mnt/ansys_inc/v212/'
    if 'DPF_PATH' not in os.environ:
        os.environ['DPF_PATH'] = '/mnt/ansys_inc/dpf/bin_v%s/Ans.dpf.core.Grpc.exe' % __version__
    if 'AWP_UNIT_TEST_FILES' not in os.environ:
        os.environ['AWP_UNIT_TEST_FILES'] = '/mnt/ansys_inc/dpf/test_files/'

from ansys.dpf.core.misc import module_exists, Report
from ansys.dpf.core.dpf_operator import Operator, Config
from ansys.dpf.core.model import Model
from ansys.dpf.core.field import Field, FieldDefinition, Dimensionnality
from ansys.dpf.core.property_field import PropertyField
from ansys.dpf.core.fields_container import FieldsContainer
from ansys.dpf.core.meshes_container import MeshesContainer
from ansys.dpf.core.scopings_container import ScopingsContainer
from ansys.dpf.core.server import (start_local_server, _global_server,
                                   connect_to_server, has_local_server)
from ansys.dpf.core.data_sources import DataSources
from ansys.dpf.core.scoping import Scoping
from ansys.dpf.core.common import types, natures, locations, shell_layers
from ansys.dpf.core import help
from ansys.dpf.core.core import BaseService, load_library, download_file, upload_file,upload_file_in_tmp_folder, upload_files_in_folder, download_files_in_folder, make_tmp_dir_server
from ansys.dpf.core.time_freq_support import TimeFreqSupport
from ansys.dpf.core.meshed_region import MeshedRegion
from ansys.dpf.core.elements import element_types
from ansys.dpf.core.result_info import ResultInfo
from ansys.dpf.core.collection import Collection
from ansys.dpf.core.workflow import Workflow
from ansys.dpf.core.cyclic_support import CyclicSupport
from ansys.dpf.core import operators
from ansys.dpf.core.fields_factory import field_from_array
from ansys.dpf.core import fields_container_factory,fields_factory, mesh_scoping_factory, time_freq_scoping_factory
from ansys.dpf.core import server
from ansys.dpf.core import check_version

# for matplotlib
# solves "QApplication: invalid style override passed, ignoring it."
os.environ['QT_STYLE_OVERRIDE'] = ''

# Setup data directory
USER_DATA_PATH = None
EXAMPLES_PATH = None
if os.environ.get('DPF_DOCKER', False):  # pragma: no cover
    # Running DPF within docker (likely for CI)
    # path must be relative to DPF directory
    #
    # assumes the following docker mount:
    # -v /tmp:/dpf/_cache
    EXAMPLES_PATH = '/tmp'
else:
    try:
        import appdirs
        USER_DATA_PATH = appdirs.user_data_dir('ansys-dpf-core')
        if not os.path.exists(USER_DATA_PATH):  # pragma: no cover
            os.makedirs(USER_DATA_PATH)

        EXAMPLES_PATH = os.path.join(USER_DATA_PATH, 'examples')
        if not os.path.exists(EXAMPLES_PATH):  # pragma: no cover
            os.makedirs(EXAMPLES_PATH)
    except:  # pragma: no cover
        pass


# Configure PyVista's ``rcParams`` for dpf
if module_exists("pyvista"):
    import pyvista as pv
    pv.rcParams['interactive'] = True
    pv.rcParams["cmap"] = "jet"
    pv.rcParams["font"]["family"] = "courier"
    pv.rcParams["title"] = "DPF"


SERVER = None

_server_instances = []



"""
Collection
===========
Contains classes associated to the DPF Collection

"""

from ansys import dpf
from ansys.grpc.dpf import collection_pb2, collection_pb2_grpc
from ansys.dpf.core.core import base_pb2
from ansys.dpf.core.common import types
from ansys.dpf.core.scoping import Scoping, scoping_pb2
from ansys.dpf.core.field import Field, field_pb2
from ansys.dpf.core.meshed_region import MeshedRegion, meshed_region_pb2
from ansys.dpf.core.time_freq_support import TimeFreqSupport
from ansys.dpf.core.errors import protect_grpc
from ansys.dpf.core import server
from ansys.dpf.core.scoping import Scoping


class Collection:
    """A class used to represent a Collection which contains
    entries ordered in labels and ids.

    Parameters
    ----------
    collection : ansys.grpc.dpf.collection_pb2.Collection, optional
        Create a collection from a Collection message.

    server : server.DPFServer, optional
        Server with channel connected to the remote or local instance. When
        ``None``, attempts to use the the global server.
        
    Examples
    --------
    >>> from ansys.dpf import core as dpf
    >>> coll = dpf.Collection(dpf.types.field)

    """

    def __init__(self, dpf_type, collection=None, server: server.DpfServer=None ):
        if server is None:
            server = dpf.core._global_server()

        self._server = server
        self._stub = self._connect()
        self._type = dpf_type
        # self.__info = None  # cached info

        if collection is None:
            request = collection_pb2.CollectionRequest()
            if hasattr(dpf_type, 'name'):
                stype = dpf_type.name
            else:
                stype = dpf_type
            request.type = base_pb2.Type.Value(stype.upper())
            self._message = self._stub.Create(request)
        elif hasattr(collection, '_message'):
            self._message = collection._message
            self._collection = collection #keep the base collection used for copy
        else:
            self._message = collection

    def set_labels(self, labels):
        """set the requested labels to scope the collection

        Parameters
        ----------
        labels : list[str], optional
            labels on which the entries will be scoped, for example:
                ['time','complex']

        """
        if len(self._info['labels'])!=0:
            print("the collection has already labels :",self._info['labels'],'deleting existing labels is  not implemented yet')
            return
        request = collection_pb2.UpdateLabelsRequest()
        request.collection.CopyFrom(self._message)
        request.labels.extend([collection_pb2.NewLabel(label=lab) for lab in labels])
        self._stub.UpdateLabels(request)

    def add_label(self, label, default_value =None):
        """Add the requested label to scope the collection

        Parameters
        ----------
        label : str
            Labels on which the entries will be scoped, for example ``'time'``.
            
        default_value : int , optional
            default value set for existing fields in the collection

        Examples
        --------
        >>> from ansys.dpf import core as dpf
        >>> coll = dpf.Collection(dpf.types.field)
        >>> coll.add_label('time')
        
        """
        request = collection_pb2.UpdateLabelsRequest()
        request.collection.CopyFrom(self._message)
        new_label = collection_pb2.NewLabel(label=label)
        if default_value is not None:
            new_label.default_value.default_value=default_value
        request.labels.extend([new_label])
        self._stub.UpdateLabels(request)

    def _get_labels(self):
        """get the labels scoping the collection

        Returns
        -------
        labels: list[str]
            labels on which the entries are scoped, for example:
                ``['time', 'complex']``
        """
        return self._info['labels']

    labels = property(_get_labels, set_labels, "labels")
    
    def has_label(self, label):
        """Check if a collection has a specified label

        Parameters
        ----------
        label: str
            Labels that must be searched, for example ``'time'``.
            
        Returns
        -------
        bool
            ``True`` if the specified value has been found in the collection. 

        Examples
        --------
        >>> from ansys.dpf import core as dpf
        >>> coll = dpf.Collection(dpf.types.field)
        >>> coll.add_label('time')
        >>> coll.has_label('time')
        True
        
        >>> coll.has_label('complex')
        False
        
        """
        return label in self.labels

    def _get_entries(self, label_space_or_index):
        """Returns the entries at a requested index or label space

        Parameters
        ----------
        label_space_or_index : dict[str,int]
            Label space of the requested entry, for example:
            ``{"time": 1, "complex": 0}`` or index of the field.

        Returns
        -------
        entries : list[Scoping], list[Field], list[MeshedRegion]
            entries corresponding to the request
        """
        request = collection_pb2.EntryRequest()
        request.collection.CopyFrom(self._message)

        if isinstance(label_space_or_index, dict):
            for key in label_space_or_index:
                request.label_space.label_space[key] = label_space_or_index[key]
        elif isinstance(label_space_or_index, int):
            request.index = label_space_or_index
        
        out = self._stub.GetEntries(request)
        list_out =[]
        for obj in out.entries :
            if obj.HasField("dpf_type"):
                if self._type == types.scoping:
                    unpacked_msg = scoping_pb2.Scoping()
                    obj.dpf_type.Unpack(unpacked_msg)
                    list_out.append(Scoping(scoping=unpacked_msg, server=self._server))
                elif self._type == types.field:
                    unpacked_msg = field_pb2.Field()
                    obj.dpf_type.Unpack(unpacked_msg)
                    list_out.append(Field(field=unpacked_msg, server=self._server))
                elif self._type == types.meshed_region:
                    unpacked_msg = meshed_region_pb2.MeshedRegion()
                    obj.dpf_type.Unpack(unpacked_msg)
                    list_out.append(MeshedRegion(mesh=unpacked_msg, server=self._server))
        if len(list_out)==0:
            list_out=None
        return list_out  
    
    
    def _get_entry(self, label_space_or_index):
        """Returns the entry at a requested index or label space

        Parameters
        ----------
        label_space_or_index : dict[str,int]
            Label space of the requested entry, for example:
            ``{"time": 1, "complex": 0}`` or index of the field.

        Returns
        -------
        entry : Scoping, Field, MeshedRegion
            entry corresponding to the request
        """
        entries = self._get_entries(label_space_or_index)
        if isinstance(entries, list):
            if len(entries)==1:
                return entries[0]
            else :
                raise KeyError(f"{label_space_or_index} has {len(entries)} entries")
        else:
            return entries
        
            
    def get_label_space(self, index):
        """Returns the label space of an entry at a requested index

        Parameters
        ----------
        index: int, optional
            Index of the entry.

        Returns
        -------
        label_space : dict(str:int)
            Scoping of the requested entry, for example:
            ``{"time": 1, "complex": 0}``
        """
        request = collection_pb2.EntryRequest()
        request.collection.CopyFrom(self._message)
        request.index = index
        out = self._stub.GetEntries(request).entries
        out = out[0].label_space.label_space
        dictOut ={}
        for key in out:
            dictOut[key]=out[key]

        return dictOut

    def get_available_ids_for_label(self, label="time"):
        """Get the IDs corresponding to the input label.


        Parameters
        ----------
        label : str
            name of the requested ids

        Returns
        -------
        ids : list[int]
            ids corresponding to the input label
        """
        ids = []
        for i in range(len(self)):
            current_scop = self.get_label_space(i)
            if label in current_scop and current_scop[label] not in ids:
                ids.append(current_scop[label])
        return ids
    
    def get_label_scoping(self, label = "time"):
        """Get the scoping corresponding to the input label. This method
        allows to get the list of available ids for a given label in the 
        Collection. 
        For example, if the label "el_type" is available in the collection,
        the get_lable_scoping method, will return the list element type ids 
        available in it. Those ids can then be used to request a given entity 
        inside the collection.
        
        Parameters
        ----------
        label: str
            name of the requested ids
        
        Returns
        -------
        scoping: Scoping
            scoping containing the ids of the input label
        """
        request = collection_pb2.LabelScopingRequest()
        request.collection.CopyFrom(self._message)
        request.label = label
        scoping_message = self._stub.GetLabelScoping(request)
        scoping = Scoping(scoping_message.label_scoping)
        return scoping

    def __getitem__(self, index):
        """Returns the entry at a requested index

        Parameters
        ----------
        key : int
            the index

        Returns
        -------
        entry : Field , Scoping
            Entry at the index corresponding to the request.
        """
        self_len = len(self)
        if index < 0:
            # convert to a positive index
            index = self_len + index

        if not self_len:
            raise IndexError('This collection contains no items')
        if index >= self_len:
            raise IndexError(f'This collection contains only {self_len} entrie(s)')

        return self._get_entries(index)[0]

    def _add_entry(self, label_space, entry):
        """Update or add the entry at a requested label space

        parameters
        ----------
        label_space : list[str,int]
            label space of the requested fields, ex : {"time":1, "complex":0}

        entry : Field or Scoping
            DPF entry to add.
        """
        request = collection_pb2.UpdateRequest()
        request.collection.CopyFrom(self._message)
        request.entry.dpf_type.Pack(entry._message)
        for key in label_space:
            request.label_space.label_space[key] = label_space[key]
        self._stub.UpdateEntry(request)
        
    def _get_time_freq_support(self):
        """
        Returns
        -------
        time_freq_support : TimeFreqSupport
        """
        request = collection_pb2.SupportRequest()
        request.collection.CopyFrom(self._message)
        request.type = base_pb2.Type.Value("TIME_FREQ_SUPPORT")
        message = self._stub.GetSupport(request)
        return TimeFreqSupport(time_freq_support=message)


    def _set_time_freq_support(self, time_freq_support):
        """
        Set the time_freq_support of a collection. 
        """
        request = collection_pb2.UpdateSupportRequest()
        request.collection.CopyFrom(self._message)
        request.time_freq_support.CopyFrom(time_freq_support._message)
        request.label = "time"
        self._stub.UpdateSupport(request)
    

    def _connect(self):
        """Connect to the grpc service"""
        return collection_pb2_grpc.CollectionServiceStub(self._server.channel)

    @property
    @protect_grpc
    def _info(self):
        """Length and labels of this container"""
        list_stub = self._stub.List(self._message)
        return {"len": list_stub.count_entries, "labels": list_stub.labels.labels}
    
    def __str__(self):
        """describe the entity
        
        Returns
        -------
        description : str
        """
        request = base_pb2.DescribeRequest()
        request.dpf_type_id = self._message.id
        return self._stub.Describe(request).description

    def __len__(self):
        """Return number of entries"""
        return self._info["len"]

    def __del__(self):
        try:
            self._stub.Delete(self._message)
        except:
            pass

    def __iter__(self):
        for i in range(len(self)):
            yield self[i]



"""
fields_factory
==============

Contains functions to make easy fields creation.
"""

from ansys.dpf import core
from ansys.dpf.core.common import natures, locations
from ansys.dpf.core import Field
from ansys.dpf.core.field import Dimensionnality
from ansys.grpc.dpf import field_pb2, field_pb2_grpc, base_pb2

import numpy as np

def field_from_array(arr):
    """Creates DPF vector or scalar field from a numpy array or a
    Python list.

    Parameters
    ----------
    arr : np.ndarray or List
        Numpy array or Python List containing either 1 or 3 dimensions.

    Returns
    -------
    field : Field
        Field constructed from numpy array.
    """
    from ansys.dpf.core import Field, natures
    arr = np.asarray(arr)

    if not np.issubdtype(arr.dtype, np.number):
        raise TypeError('Array must be a numeric type')

    shp_err = ValueError('Array must be either contain 1 dimension or '
                         '2 dimensions with three components.')
    if arr.ndim == 1:
        nature = natures.scalar
    elif arr.ndim == 2:
        if arr.shape[1] == 1:
            arr = arr.ravel()
            nature = natures.scalar
        elif arr.shape[1] == 3:
            nature = natures.vector
        elif arr.shape[1] == 6:
            nature = natures.symmatrix
        else:
            raise shp_err
    else:
        raise shp_err

    n_entities = arr.shape[0]
    field = Field(nentities=n_entities, nature=nature)
    field.data = arr
    field.scoping.ids = np.arange(1, n_entities + 1)
    return field

def create_matrix_field(num_entities, num_lines, num_col, location = locations.nodal, server=None):
    """Helper function to create a specific ``ansys.dpf.core.Field``.
    The returned field will contain entities that have matrix format. 
    This is a "reserve" mechanism, not a resize one. This means that you
    need to append data to grow up the size of your field. 

    Parameters
    ----------
    num_entities : int
        Number of entities reserved
        
    num_lines : int
        Number of matrix lines 
    
    num_col : int
        Number of matrix columns

    location : str, optional
        Location of the field. Default: ``"Nodal"``. For example:

        - ``"Nodal"``
        - ``"Elemental"``
        - ansys.dpf.core.natures.elemental_nodal
    
    server : server.DPFServer, optional
        Server with channel connected to the remote or local instance. When
        ``None``, attempts to use the the global server.   
        
    Returns
    -------
    field : Field
        Dpf field at the requested format.

    Examples
    --------
    Create field containing 3 matrix entities of a col*lines = 2*5 size with
    nodal location (default). 

    >>> from ansys.dpf.core import fields_factory
    >>> field = fields_factory.create_matrix_field(3, 5, 2)
    
    """
    return _create_field(server, natures.matrix, num_entities, location, num_col, num_lines)

def create_3d_vector_field(num_entities, location = locations.nodal, server=None):
    """Helper function to create a specific ``ansys.dpf.core.Field``.
    The returned field will contain entities that have 3d vector format. 
    This is a "reserve" mechanism, not a resize one. This means that you
    need to append data to grow up the size of your field. 

    Parameters
    ----------
    num_entities : int
        Number of entities reserved

    location : str, optional
        Location of the field. Default: ``"Nodal"``. For example:

        - ``"Nodal"``
        - ``"Elemental"``
        - ansys.dpf.core.natures.elemental_nodal
    
    server : server.DPFServer, optional
        Server with channel connected to the remote or local instance. When
        ``None``, attempts to use the the global server.   
        
    Returns
    -------
    field : Field
        Dpf field at the requested format.

    Examples
    --------
    Create field containing 4 3d vector entities with nodal location (default). 

    >>> from ansys.dpf.core import fields_factory
    >>> field = fields_factory.create_3d_vector_field(4)
    
    """
    return _create_field(server, natures.vector, num_entities, location)

def create_tensor_field(num_entities, location = locations.nodal, server=None):
    """Helper function to create a specific ``ansys.dpf.core.Field``.
    The returned field will contain entities that have 3*3 format. 
    This is a "reserve" mechanism, not a resize one. This means that you
    need to append data to grow up the size of your field. 

    Parameters
    ----------
    num_entities : int
        Number of entities reserved

    location : str, optional
        Location of the field. Default: ``"Nodal"``. For example:

        - ``"Nodal"``
        - ``"Elemental"``
        - ansys.dpf.core.natures.elemental_nodal
    
    server : server.DPFServer, optional
        Server with channel connected to the remote or local instance. When
        ``None``, attempts to use the the global server.   
        
    Returns
    -------
    field : Field
        Dpf field at the requested format.

    Examples
    --------
    Create field containing 4 tensor entities with nodal location (default). 

    >>> from ansys.dpf.core import fields_factory
    >>> field = fields_factory.create_tensor_field(4)
    
    """
    return _create_field(server, natures.symmatrix, num_entities, location)

def create_scalar_field(num_entities, location = locations.nodal, server=None):
    """Helper function to create a specific ``ansys.dpf.core.Field``.
    The returned field will contain entities that are scalar. 
    This is a "reserve" mechanism, not a resize one. This means that you
    need to append data to grow up the size of your field. 

    Parameters
    ----------
    num_entities : int
        Number of entities reserved

    location : str, optional
        Location of the field. Default: ``"Nodal"``. For example:

        - ``"Nodal"``
        - ``"Elemental"``
        - ansys.dpf.core.natures.elemental_nodal
    
    server : server.DPFServer, optional
        Server with channel connected to the remote or local instance. When
        ``None``, attempts to use the the global server.   
        
    Returns
    -------
    field : Field
        Dpf field at the requested format.

    Examples
    --------
    Create field containing 4 scalars with nodal location (default). 

    >>> from ansys.dpf.core import fields_factory
    >>> field = fields_factory.create_scalar_field(4)
    
    """
    return _create_field(server, natures.scalar, num_entities, location)

def create_vector_field(num_entities, num_comp, location = locations.nodal, server=None):
    """Helper function to create a specific ``ansys.dpf.core.Field``.
    The returned field will contain entities that have vector format. 
    This is a "reserve" mechanism, not a resize one. This means that you
    need to append data to grow up the size of your field. 

    Parameters
    ----------
    num_entities : int
        Number of entities reserved
        
    num_comp : int
        Number of vector's components 

    location : str, optional
        Location of the field. Default: ``"Nodal"``. For example:

        - ``"Nodal"``
        - ``"Elemental"``
        - ansys.dpf.core.natures.elemental_nodal
    
    server : server.DPFServer, optional
        Server with channel connected to the remote or local instance. When
        ``None``, attempts to use the the global server.   
        
    Returns
    -------
    field : Field
        Dpf field at the requested format.

    Examples
    --------
    Create field containing 3 vectors entities of 5 components each with
    nodal location (default). 

    >>> from ansys.dpf.core import fields_factory
    >>> field = fields_factory.create_vector_field(3, 5)
    
    """
    return _create_field(server, natures.vector, num_entities, location, ncomp_n = num_comp)

def _connect(server):
    """Connect to the grpc instance"""
    if server is None:
        server = core._global_server()
    stub = field_pb2_grpc.FieldServiceStub(server.channel)
    return stub

def _create_field(server, nature, nentities, location = locations.nodal, 
                  ncomp_n = 0, ncomp_m = 0):
    """Private helper function to create a specific ``ansys.dpf.core.Field``.
    This is a "reserve" mechanism, not a resize one. This means that you
    need to append data to grow up the size of your field. 

    Parameters
    ----------
    server : server.DPFServer, optional
        Server with channel connected to the remote or local instance. When
        ``None``, attempts to use the the global server. 
    
    snature : str 
        Defines the nature of the field entity data. For example: 
            
        - ansys.dpf.core.natures.matrix
        - ansys.dpf.core.natures.scalar
        
        
    num_entities : int
        Number of entities reserved

    location : str optional
        Location of the field.  For example:

        - ``"Nodal"``
        - ``"Elemental"``
        - ansys.dpf.core.locations.elemental_nodal    
            
    ncomp_n : int
        Number of lines 
    
    ncomp_m : int
        Number of columns
        
    Returns
    -------
    field : Field
        Dpf field at the requested format."""
    # ncomp_n is number of column components
    # ncomp_m is number of line components
    # connect to grpc
    stub = _connect(server)
    # set nature
    if hasattr(nature, 'name'):
        snature = nature.name
    else:
        snature = nature
        
    if snature == natures.vector.name:
        elem_data_size = 3
    elif snature == natures.symmatrix.name:
        elem_data_size = 6
    elif snature == natures.scalar.name:
        elem_data_size=1
    elif snature == natures.matrix.name: 
        elem_data_size = ncomp_n * ncomp_m
    else:
        elem_data_size = ncomp_n 
    if (ncomp_n != 0 and ncomp_m != 0):
        dimensionality = Dimensionnality([ncomp_n, ncomp_m], nature)
    elif (ncomp_n != 0 and ncomp_m == 0):
        dimensionality = Dimensionnality([ncomp_n], nature)
    else: 
        dimensionality = None        
    # set request
    request = field_pb2.FieldRequest()
    nature = base_pb2.Nature.Value(snature.upper())
    request.nature = nature
    request.location.location = location
    request.size.scoping_size = nentities
    if dimensionality is not None:
        request.dimensionality.CopyFrom(dimensionality._parse_dim_to_message())
    request.size.data_size = nentities*elem_data_size
    # get field
    message = stub.Create(request)
    field = Field(field = message, server = server)
    return field


"""
skin
====
"""
from ansys.dpf.core.dpf_operator import Operator
from ansys.dpf.core.inputs import Input, _Inputs
from ansys.dpf.core.outputs import Output, _Outputs, _modify_output_spec_with_one_type
from ansys.dpf.core.operators.specification import PinSpecification, Specification

"""Operators from meshOperatorsCore plugin, from "mesh" category
"""

class skin(Operator):
    """Extracts a skin of the mesh (2D elements) in a new meshed region. Material id of initial elements are propagated to their facets.

      available inputs:
        - mesh (MeshedRegion)
        - mesh_scoping (Scoping) (optional)

      available outputs:
        - mesh (MeshedRegion)
        - nodes_mesh_scoping (Scoping)
        - map_new_elements_to_old ()
        - property_field_new_elements_to_old (PropertyField)

      Examples
      --------
      >>> from ansys.dpf import core as dpf

      >>> # Instantiate operator
      >>> op = dpf.operators.mesh.skin()

      >>> # Make input connections
      >>> my_mesh = dpf.MeshedRegion()
      >>> op.inputs.mesh.connect(my_mesh)
      >>> my_mesh_scoping = dpf.Scoping()
      >>> op.inputs.mesh_scoping.connect(my_mesh_scoping)

      >>> # Instantiate operator and connect inputs in one line
      >>> op = dpf.operators.mesh.skin(mesh=my_mesh,mesh_scoping=my_mesh_scoping)

      >>> # Get output data
      >>> result_mesh = op.outputs.mesh()
      >>> result_nodes_mesh_scoping = op.outputs.nodes_mesh_scoping()
      >>> result_map_new_elements_to_old = op.outputs.map_new_elements_to_old()
      >>> result_property_field_new_elements_to_old = op.outputs.property_field_new_elements_to_old()"""
    def __init__(self, mesh=None, mesh_scoping=None, config=None, server=None):
        super().__init__(name="meshed_skin_sector", config = config, server = server)
        self._inputs = InputsSkin(self)
        self._outputs = OutputsSkin(self)
        if mesh !=None:
            self.inputs.mesh.connect(mesh)
        if mesh_scoping !=None:
            self.inputs.mesh_scoping.connect(mesh_scoping)

    @staticmethod
    def _spec():
        spec = Specification(description="""Extracts a skin of the mesh (2D elements) in a new meshed region. Material id of initial elements are propagated to their facets.""",
                             map_input_pin_spec={
                                 0 : PinSpecification(name = "mesh", type_names=["abstract_meshed_region"], optional=False, document=""""""), 
                                 1 : PinSpecification(name = "mesh_scoping", type_names=["scoping"], optional=True, document="""""")},
                             map_output_pin_spec={
                                 0 : PinSpecification(name = "mesh", type_names=["abstract_meshed_region"], optional=False, document="""skin meshed region with facets and facets_to_ele property fields"""), 
                                 1 : PinSpecification(name = "nodes_mesh_scoping", type_names=["scoping"], optional=False, document=""""""), 
                                 2 : PinSpecification(name = "map_new_elements_to_old", type_names=[], optional=False, document=""""""), 
                                 3 : PinSpecification(name = "property_field_new_elements_to_old", type_names=["property_field"], optional=False, document="""""")})
        return spec


    @staticmethod
    def default_config():
        return Operator.default_config(name = "meshed_skin_sector")

    @property
    def inputs(self):
        """Enables to connect inputs to the operator

        Returns
        --------
        inputs : InputsSkin 
        """
        return super().inputs


    @property
    def outputs(self):
        """Enables to get outputs of the operator by evaluationg it

        Returns
        --------
        outputs : OutputsSkin 
        """
        return super().outputs


#internal name: meshed_skin_sector
#scripting name: skin
class InputsSkin(_Inputs):
    """Intermediate class used to connect user inputs to skin operator

      Examples
      --------
      >>> from ansys.dpf import core as dpf

      >>> op = dpf.operators.mesh.skin()
      >>> my_mesh = dpf.MeshedRegion()
      >>> op.inputs.mesh.connect(my_mesh)
      >>> my_mesh_scoping = dpf.Scoping()
      >>> op.inputs.mesh_scoping.connect(my_mesh_scoping)
    """
    def __init__(self, op: Operator):
        super().__init__(skin._spec().inputs, op)
        self._mesh = Input(skin._spec().input_pin(0), 0, op, -1) 
        self._inputs.append(self._mesh)
        self._mesh_scoping = Input(skin._spec().input_pin(1), 1, op, -1) 
        self._inputs.append(self._mesh_scoping)

    @property
    def mesh(self):
        """Allows to connect mesh input to the operator

        Parameters
        ----------
        my_mesh : MeshedRegion, 

        Examples
        --------
        >>> from ansys.dpf import core as dpf

        >>> op = dpf.operators.mesh.skin()
        >>> op.inputs.mesh.connect(my_mesh)
        >>> #or
        >>> op.inputs.mesh(my_mesh)

        """
        return self._mesh

    @property
    def mesh_scoping(self):
        """Allows to connect mesh_scoping input to the operator

        Parameters
        ----------
        my_mesh_scoping : Scoping, 

        Examples
        --------
        >>> from ansys.dpf import core as dpf

        >>> op = dpf.operators.mesh.skin()
        >>> op.inputs.mesh_scoping.connect(my_mesh_scoping)
        >>> #or
        >>> op.inputs.mesh_scoping(my_mesh_scoping)

        """
        return self._mesh_scoping

class OutputsSkin(_Outputs):
    """Intermediate class used to get outputs from skin operator
      Examples
      --------
      >>> from ansys.dpf import core as dpf

      >>> op = dpf.operators.mesh.skin()
      >>> # Connect inputs : op.inputs. ...
      >>> result_mesh = op.outputs.mesh()
      >>> result_nodes_mesh_scoping = op.outputs.nodes_mesh_scoping()
    """
    def __init__(self, op: Operator):
        super().__init__(skin._spec().outputs, op)
        self._mesh = Output(skin._spec().output_pin(0), 0, op) 
        self._outputs.append(self._mesh)
        self._nodes_mesh_scoping = Output(skin._spec().output_pin(1), 1, op) 
        self._outputs.append(self._nodes_mesh_scoping)
        pass 

    @property
    def mesh(self):
        """Allows to get mesh output of the operator


        - pindoc: skin meshed region with facets and facets_to_ele property fields

        Returns
        ----------
        my_mesh : MeshedRegion, 

        Examples
        --------
        >>> from ansys.dpf import core as dpf

        >>> op = dpf.operators.mesh.skin()
        >>> # Connect inputs : op.inputs. ...
        >>> result_mesh = op.outputs.mesh() 
        """
        return self._mesh

    @property
    def nodes_mesh_scoping(self):
        """Allows to get nodes_mesh_scoping output of the operator


        Returns
        ----------
        my_nodes_mesh_scoping : Scoping, 

        Examples
        --------
        >>> from ansys.dpf import core as dpf

        >>> op = dpf.operators.mesh.skin()
        >>> # Connect inputs : op.inputs. ...
        >>> result_nodes_mesh_scoping = op.outputs.nodes_mesh_scoping() 
        """
        return self._nodes_mesh_scoping




"""
Core
====
"""
import os
import logging
import time
import weakref
import pathlib

import grpc

from ansys.grpc.dpf import base_pb2, base_pb2_grpc
from ansys.dpf.core.errors import protect_grpc
from ansys.dpf.core import server as serverlib
from ansys.dpf.core.misc import DEFAULT_FILE_CHUNK_SIZE
from ansys.dpf.core.common import _common_progress_bar

LOG = logging.getLogger(__name__)
LOG.setLevel('DEBUG')

if 'DPF_CONFIGURATION' in os.environ:
    CONFIGURATION = os.environ['DPF_CONFIGURATION']
else:
    CONFIGURATION = 'release'
    

def load_library(filename, name='', symbol="LoadOperators", server=None):
    """Dynamically load an operators library for dpf.core.
    Code containing this library's operators is generated in
    ansys.dpf.core.operators

    Parameters
    ----------
    filename : str
        Filename of the operator library.

    name : str, optional
        Library name.  Probably optional
        
    server : server.DPFServer, optional
        Server with channel connected to the remote or local instance. When
        ``None``, attempts to use the the global server.

    Examples
    --------
    Load the mesh operators for Windows (for Linux, just use 
    'libmeshOperatorsCore.so' instead of 'meshOperatorsCore.dll')

    >>> from ansys.dpf import core as dpf
    >>> # dpf.load_library('meshOperatorsCore.dll', 'mesh_operators')
    
    """
    base = BaseService(server, load_operators=False)    
    base.load_library(filename, name, symbol)
    return name+ " successfully loaded";

def upload_file_in_tmp_folder(file_path, new_file_name=None, server=None):
    """Upload a file from the client to the server in a temporary folder 
    deleted when the server is shutdown
    
    Parameters
    ----------
    file_path : str
        file path on the client side to upload
        
    new_file_name : str, optional
        name to give to the file server side, 
        if no name is specified, the same name as the input file is given
        
    server : server.DPFServer, optional
        Server with channel connected to the remote or local instance. When
        ``None``, attempts to use the the global server.
        
    Notes
    -----
    Print a progress bar
        
    Returns
    -------
      server_file_path : str
           path generated server side
          
    Examples
    --------
    >>> from ansys.dpf import core as dpf
    >>> from ansys.dpf.core import examples
    >>> file_path = dpf.upload_file_in_tmp_folder(examples.static_rst)
    
    """
    base = BaseService(server, load_operators=False)    
    return base.upload_file_in_tmp_folder(file_path, new_file_name)

def upload_files_in_folder(to_server_folder_path, client_folder_path, specific_extension=None, server=None):
    """Upload all the files from a folder of the client 
    to the target server folder path.
    
    Parameters
    ----------
    to_server_folder_path : str
        folder path target where will be uploaded files on the server side
        
    client_folder_path: str
        folder path where the files that must be uploaded are located 
        on client side
        
    specific_extension (optional) : str
        copies only the files with the given extension
        
    server : server.DPFServer, optional
        Server with channel connected to the remote or local instance. When
        ``None``, attempts to use the the global server.
        
    Returns
    -------
    paths : list of str
        new file paths server side
    """    
    base = BaseService(server, load_operators=False)    
    return base.upload_files_in_folder(to_server_folder_path, client_folder_path, specific_extension)

def download_file(server_file_path, to_client_file_path, server=None):
    """Download a file from the server to the target client file path
    
    Parameters
    ----------
    server_file_path : str
        file path to download on the server side
        
    to_client_file_path: str
        file path target where the file will be located client side
        
    server : server.DPFServer, optional
        Server with channel connected to the remote or local instance. When
        ``None``, attempts to use the the global server.
        
    Notes
    -----
    Print a progress bar
    
    Examples
    --------
    >>> from ansys.dpf import core as dpf
    >>> from ansys.dpf.core import examples
    >>> import os
    >>> file_path = dpf.upload_file_in_tmp_folder(examples.static_rst)
    >>> dpf.download_file(file_path, examples.static_rst)
    
    """
    base = BaseService(server, load_operators=False)    
    return base.download_file(server_file_path, to_client_file_path)

def download_files_in_folder(server_folder_path, to_client_folder_path, specific_extension=None, server=None):
    """Download all the files from a folder of the server 
    to the target client folder path
    
    Parameters
    ----------
    server_folder_path : str
        folder path to download on the server side
        
    to_client_folder_path: str
        folder path target where the files will be located client side      
        
    specific_extension (optional) : str
        copies only the files with the given extension
        
    server : server.DPFServer, optional
        Server with channel connected to the remote or local instance. When
        ``None``, attempts to use the the global server.
        
    Notes
    -----
    Print a progress bar
        
    Returns
    -------
    paths : list of str
        new file paths client side
    """
    base = BaseService(server, load_operators=False)    
    return base.download_files_in_folder(server_folder_path, to_client_folder_path,specific_extension)

        
def upload_file(file_path, to_server_file_path, server=None):
    """Upload a file from the client to the target server file path
    
    Parameters
    ----------
    file_path : str
        file path on the client side to upload
        
    to_server_file_path: str
        file path target where the file will be located server side
    
    server : server.DPFServer, optional
        Server with channel connected to the remote or local instance. When
        ``None``, attempts to use the the global server.
        
    Notes
    -----
    Print a progress bar
        
    Returns
    -------
    server_file_path : str
        path generated server side
    """
    base = BaseService(server, load_operators=False)    
    return base.upload_file(file_path, to_server_file_path)

def make_tmp_dir_server(server=None):
    """Create a temporary folder server side. Only one temporary folder can be created 
    by server instance.
    The folder will be deleted when the server is stopped.
    
    Parameters
    ----------
    server : server.DPFServer, optional
        Server with channel connected to the remote or local instance. When
        ``None``, attempts to use the the global server.
        
    Returns
    -------
    path : str
        path to the temporary dir
    """
    base = BaseService(server, load_operators=False)    
    return base.make_tmp_dir_server()

def _description(dpf_entity_message, server=None):
    """Ask the server to describe the entity in input    
    
    Parameters
    ----------
    dpf_entity_message : core.Operator._message, core.Workflow._message, core.Scoping._message, core.Field._message,
    core.FieldContainer._message, core.MeshedRegion._message...
    
    server : server.DPFServer, optional
        Server with channel connected to the remote or local instance. When
        ``None``, attempts to use the the global server.
        
    Returns
    -------
       description : str
    """
    return BaseService(server, load_operators=False)._description(dpf_entity_message)    

class BaseService():
    """The Base Service class allows to make generic requests to dpf's server.
    For example, information about the server can be requested, 
    uploading/downloading file from and to the server can be done,
    new operators plugins can be loaded...
    Most of the request done by the BaseService class are wrapped by 
    functions.

    Parameters
    ----------
    server : server.DPFServer, optional
        Server with channel connected to the remote or local instance. When
        ``None``, attempts to use the the global server.

    timeout : float, optional
        Fails when a connection takes longer than ``timeout`` seconds
        to initialize.
    
    load_operators : bool, optional
        Automatically load the math operators

    Examples
    --------
    Connect to an existing DPF server
    >>> from ansys.dpf import core as dpf
    >>> #server = dpf.connect_to_server(ip='127.0.0.1', port = 50054, as_global=False)
    >>> #base = dpf.BaseService(server=server)
    
    """

    def __init__(self, server=None, load_operators=True, timeout=5):
        """Initialize base service"""

        if server is None:
            server = serverlib._global_server()

        self._server = weakref.ref(server)
        self._stub = self._connect(timeout)

    def _connect(self, timeout=5):
        """Connect to dpf service within a given timeout"""
        stub = base_pb2_grpc.BaseServiceStub(self._server().channel)

        # verify connected
        if timeout is not None:
            state = grpc.channel_ready_future(self._server().channel)
            tstart = time.time()
            while (time.time() - tstart) < timeout and not state._matured:
                time.sleep(0.01)

            if not state._matured:
                raise IOError(f'Unable to connect to DPF instance at {self._server()._input_ip} {self._server()._input_port}')

        return stub
    
    
    def make_tmp_dir_server(self):
        """Create a temporary folder server side. Only one temporary folder can be created 
        by server instance.
        The folder will be deleted when the server is stopped.
        
        Returns
        -------
        path : str
            path to the temporary dir
        """
        request = base_pb2.Empty()
        return self._stub.CreateTmpDir(request).server_file_path


    def load_library(self, filename, name='', symbol="LoadOperators"):
        """Dynamically load an operators library for dpf.core.
        Code containing this library's operators is generated in
        ansys.dpf.core.operators

        Parameters
        ----------
        filename : str
            Filename of the operator library.

        name : str, optional
            Library name.  Probably optional

        Examples
        --------
        Load the mesh operators for Windows (for Linux, just use 
        'libmeshOperatorsCore.so' instead of 'meshOperatorsCore.dll')

        >>> from ansys.dpf import core as dpf
        >>> base = dpf.BaseService()
        >>> # base.load_library('meshOperatorsCore.dll', 'mesh_operators')
        
        """
        request = base_pb2.PluginRequest()
        request.name = name
        request.dllPath = filename
        request.symbol = symbol
        try:
            self._stub.Load(request)
        except Exception as e:
            raise IOError(f'Unable to load library "{filename}". File may not exist or'
                          f' is missing dependencies:\n{str(e)}')
        
        #TODO: fix code generation upload posix
        import os
        if os.name != 'posix':
            local_dir =os.path.dirname(os.path.abspath(__file__))
            LOCAL_PATH = os.path.join(local_dir, "operators")
            
            #send local generated code
            TARGET_PATH = self.make_tmp_dir_server()
            self.upload_files_in_folder(TARGET_PATH, LOCAL_PATH, "py")
                
            #generate code     
            from ansys.dpf.core.dpf_operator import Operator
            code_gen = Operator("python_generator")
            code_gen.connect(1,TARGET_PATH)
            code_gen.connect(0, filename)
            code_gen.connect(2, False)
            code_gen.run()    
            
            self.download_files_in_folder(TARGET_PATH, LOCAL_PATH,"py")
            
    
    @property
    def server_info(self):
        """Send the request for server information and keep 
           the info into a dictionary
           
           Returns
           -------
           info : dictionary
               dictionary with "server_ip", "server_port", "server_process_id"
               "server_version" keys
        """
        request = base_pb2.ServerInfoRequest()
        try:
            response = self._stub.GetServerInfo(request)
        except Exception as e:
            raise IOError(f'Unable to recover information from the server:\n{str(e)}')
        out = {"server_ip":response.ip, "server_port":response.port, "server_process_id":response.processId,
               "server_version": str(response.majorVersion) +"."+str(response.minorVersion)}
        return out
    
    
    def _description(self, dpf_entity_message):
        """Ask the server to describe the entity in input
        
        Parameters
        ----------
        dpf_entity_message : core.Operator._message, core.Workflow._message, core.Scoping._message, core.Field._message,
        core.FieldContainer._message, core.MeshedRegion._message...
        
        Returns
        -------
           description : str
        """
        request = base_pb2.DescribeRequest()
        request.dpf_type_id = dpf_entity_message.id
        return self._stub.Describe(request).description
    
    
    def _get_separator(self, path):
        s1 = len(path.split('\\')) 
        s2 = len(path.split('/')) 
        if s2 > s1:
            # Linux case
            separator = '/'
        elif s1 > s2:
            # Windows case
            separator = "\\"
        return separator
    
    
    @protect_grpc
    def download_file(self, server_file_path, to_client_file_path):
        """Download a file from the server to the target client file path
        
        Notes
        -----
        Print a progress bar
        
        Parameters
        ----------
        server_file_path : str
            file path to dowload on the server side
            
        to_client_file_path: str
            file path target where the file will be located client side
        """
        request = base_pb2.DownloadFileRequest()
        request.server_file_path = server_file_path
        chunks = self._stub.DownloadFile(request)
        bar =_common_progress_bar("Downloading...", unit="KB")
        bar.start()
        i = 0
        with open(to_client_file_path, 'wb') as f:
            for chunk in chunks:
                f.write(chunk.data.data)
                i += len(chunk.data.data)*1E-3
                bar.update(i)
        bar.finish()
                
    @protect_grpc
    def download_files_in_folder(self,server_folder_path, to_client_folder_path, specific_extension=None):
        """Download all the files from a folder of the server 
        to the target client folder path
        
        Parameters
        ----------
        server_folder_path : str
            folder path to download on the server side
            
        to_client_folder_path: str
            folder path target where the files will be located client side  
            
        specific_extension (optional) : str
            copies only the files with the given extension
                    
        Notes
        -----
        Print a progress bar
    
            
        Returns
        -------
        paths : list of str
            new file paths client side
        """
        request = base_pb2.DownloadFileRequest()
        request.server_file_path = server_folder_path
        chunks = self._stub.DownloadFile(request)
        
        num_files = 1
        if chunks.initial_metadata()[0].key == u"num_files":
            num_files = int(chunks.initial_metadata()[0].value)
            
        bar =_common_progress_bar("Downloading...", unit="files", tot_size = num_files)
        bar.start()
        
        server_path =""
        
        import ntpath
        client_paths=[]
        f = None
        for chunk in chunks:
            if chunk.data.server_file_path != server_path :
                server_path = chunk.data.server_file_path
                if specific_extension == None or pathlib.Path(server_path).suffix == "."+specific_extension : 
                    separator = self._get_separator(server_path)
                    server_subpath = server_path.replace(server_folder_path + separator, '')
                    subdir = ''
                    split = server_subpath.split(separator)
                    n = len(split)
                    i = 0
                    to_client_folder_path_copy = to_client_folder_path
                    if n > 1:
                        while i < (n - 1):
                            subdir = split[i]
                            subdir_path = os.path.join(to_client_folder_path_copy, subdir)
                            if not os.path.exists(subdir_path):
                                os.mkdir(subdir_path)
                            to_client_folder_path_copy = subdir_path
                            i += 1
                    cient_path = os.path.join(to_client_folder_path_copy, ntpath.basename(server_path))
                    client_paths.append(cient_path)
                    f = open(cient_path, 'wb')
                    try:
                        bar.update(len(client_paths))
                    except:
                        pass
                else:
                    f = None
            if f is not None:
                f.write(chunk.data.data)
        try:
            bar.finish()
        except:
            pass
        return client_paths
    
    @protect_grpc
    def upload_files_in_folder(self, to_server_folder_path, client_folder_path, specific_extension=None):
        """Upload all the files from a folder of the client 
        to the target server folder path.
        
        Parameters
        ----------
        to_server_folder_path : str
            folder path target where will be uploaded files on the server side
            
        client_folder_path: str
            folder path where the files that must be uploaded are located 
            on client side
            
        specific_extension (optional) : str
            copies only the files with the given extension
            
        Returns
        -------
        paths : list of str
            new file paths server side
        """              
        server_paths = []
        for root, subdirectories, files in os.walk(client_folder_path):
            for subdirectory in subdirectories:
                subdir = os.path.join(root, subdirectory)
                for filename in os.listdir(subdir):
                    f = os.path.join(subdir, filename)
                    server_paths = self._upload_and_get_server_path(specific_extension, f, filename, server_paths, to_server_folder_path, subdirectory)
            for file in files:
                f = os.path.join(root, file)
                server_paths = self._upload_and_get_server_path(specific_extension, f, file, server_paths, to_server_folder_path)
            break
        return server_paths
    
    def _upload_and_get_server_path(self, specific_extension, f, filename, server_paths, to_server_folder_path, subdirectory = None):
        separator = self._get_separator(to_server_folder_path)
                
        if subdirectory is not None:
            to_server_file_path = to_server_folder_path + separator + subdirectory + separator + filename
        else:
            to_server_file_path = to_server_folder_path + separator + filename
        if ((specific_extension is not None) and (f.endswith(specific_extension))) or (specific_extension is None):
            server_path = self._stub.UploadFile(self.__file_chunk_yielder(file_path=f, to_server_file_path=to_server_file_path)).server_file_path
            server_paths.append(server_path)
        return server_paths
                

    @protect_grpc
    def upload_file(self, file_path, to_server_file_path):
        """Upload a file from the client to the target server file path
        
        Parameters
        ----------
        file_path : str
            file path on the client side to upload
            
        to_server_file_path: str
            file path target where the file will be located server side    
        
        Notes
        -----
        Print a progress bar
            
        Returns
        -------
           server_file_path : str
               path generated server side
        """
        if os.stat(file_path).st_size==0:
            raise ValueError(file_path+" is empty")
        return self._stub.UploadFile(self.__file_chunk_yielder(file_path, to_server_file_path)).server_file_path
        
    @protect_grpc
    def upload_file_in_tmp_folder(self, file_path, new_file_name=None):
        """Upload a file from the client to the server in a temporary folder 
        deleted when the server is shutdown
        
        Parameters
        ----------
        file_path : str
            file path on the client side to upload
            
        new_file_name : str, optional
            name to give to the file server side, 
            if no name is specified, the same name as the input file is given    
        
        Notes
        -----
        Print a progress bar
            
        Returns
        -------
           server_file_path : str
               path generated server side
        """
        if new_file_name:
            file_name = new_file_name
        else:
            file_name = os.path.basename(file_path)
        if os.stat(file_path).st_size==0:
            raise ValueError(file_path+" is empty")
        return self._stub.UploadFile(self.__file_chunk_yielder(file_path=file_path, to_server_file_path=file_name, use_tmp_dir=True)).server_file_path
    
    def _prepare_shutdown(self):
        self._stub.PrepareShutdown(base_pb2.Empty())
        
                
    def __file_chunk_yielder(self,file_path, to_server_file_path, use_tmp_dir=False):
        request = base_pb2.UploadFileRequest()
        request.server_file_path = to_server_file_path
        request.use_temp_dir=use_tmp_dir
        
        tot_size = os.path.getsize(file_path)*1E-3
        
        need_progress_bar = tot_size>10000
        if need_progress_bar:
            bar =_common_progress_bar("Uploading...", "KB", tot_size)
            bar.start()
        i=0
        with open(file_path, 'rb') as f:
            while True:
                piece = f.read(DEFAULT_FILE_CHUNK_SIZE)
                if len(piece)==0:
                    break
                request.data.data = piece
                yield request
                i+=len(piece)*1E-3
                if need_progress_bar:
                    try:
                        bar.update(min(i,tot_size))
                    except :
                        pass
        
        if need_progress_bar:
            try:
                bar.finish()
            except :
                pass
                
                

        
        


"""
MeshedRegion
============
"""
from ansys import dpf
from ansys.grpc.dpf import meshed_region_pb2, meshed_region_pb2_grpc
from ansys.dpf.core import scoping
from ansys.dpf.core.common import locations, types
from ansys.dpf.core.plotter import Plotter as _DpfPlotter
from ansys.dpf.core.errors import protect_grpc
from ansys.dpf.core.nodes import Nodes
from ansys.dpf.core.elements import Elements, element_types
from ansys.dpf.core.check_version import server_meet_version


class MeshedRegion:
    """A class used to represent a Mesh from DPF.

    Parameters
    ----------
    num_nodes : int
        number of nodes to reserve for mesh creation
    
    num_elements : int
        number of elements to reserve for mesh creation
        
    mesh : ansys.grpc.dpf.meshed_region_pb2.MeshedRegion
    
    server : server.DPFServer, optional
        Server with channel connected to the remote or local instance. When
        ``None``, attempts to use the the global server.

    Attributes
    ----------
    nodes : Nodes
        Entity containing all the nodal properties

    elements : Elements
        Entity containing all the elemental properties

    Examples
    --------
    Extract a meshed region from a model.

    >>> import ansys.dpf.core as dpf
    >>> from ansys.dpf.core import examples
    >>> model = dpf.Model(examples.static_rst)
    >>> meshed_region = model.metadata.meshed_region
    
    Create a meshed region from scratch (line with 3 beam elements)
    
    >>> import ansys.dpf.core as dpf
    >>> meshed_region = dpf.MeshedRegion(num_nodes=4,num_elements=3)
    >>> i=0
    >>> for node in meshed_region.nodes.add_nodes(4):
    ...     node.id = i+1
    ...     node.coordinates = [float(i), float(i), 0.0]
    ...     i=i+1
    >>> i=0
    >>> for element in meshed_region.elements.add_elements(3):
    ...     element.id=i+1
    ...     element.connectivity = [i, i+1]
    ...     element.is_beam=True #or is_solid, is_beam, is_point
    ...     i=i+1
    >>> meshed_region.elements.add_beam_element(id=4,connectivity=[3,0])
    
    """

    def __init__(self, num_nodes=None, num_elements=None, mesh=None, server=None):

        if server is None:
            server = dpf.core._global_server()

        self._server = server
        self._stub = self._connect()

        if isinstance(mesh, MeshedRegion):
            self._message = mesh._mesh
        elif isinstance(mesh, meshed_region_pb2.MeshedRegion):
            self._message = mesh
        elif mesh==None:
            self.__send_init_request(num_nodes,num_elements)       
        else: #support_pb2.Support
            self._message = meshed_region_pb2.MeshedRegion()
            self._message.id = mesh.id
        

        self._full_grid = None
        self._elements = None
        self._nodes = None

    def _get_scoping(self, loc=locations.nodal):
        """
        Parameters
        ----------
        loc : str or ansys.dpf.core.common.locations, optional
            location of the requested scoping ("Nodal", "Elemental"...)

        Returns
        -------
        scoping : Scoping
            ids of the elements or nodes of the mesh
        """
        request = meshed_region_pb2.GetScopingRequest(mesh=self._message)
        request.loc.location = loc
        out = self._stub.GetScoping(request)
        return scoping.Scoping(scoping=out, server=self._server)

    @property
    def elements(self):
        """Gives access to all the elemental properties of the mesh:
            connectivity, element types...

        Returns
        -------
        elements : Elements
            Elements belonging to this meshed region.

        Examples
        --------
        >>> import ansys.dpf.core as dpf
        >>> from ansys.dpf.core import examples
        >>> model = dpf.Model(examples.static_rst)
        >>> meshed_region = model.metadata.meshed_region
        >>> elements = meshed_region.elements
        >>> print(elements)
        DPF Elements object with 8 elements
        
        """
        if self._elements is None:
            self._elements = Elements(self)
        return self._elements

    @property
    def nodes(self):
        """Gives access to all the nodal properties of the mesh:
            nodes coordinates, nodal connectivity...
            
        Returns
        -------
        nodes : Nodes
            Nodes belonging to this meshed region

        Examples
        --------
        >>> import ansys.dpf.core as dpf
        >>> from ansys.dpf.core import examples
        >>> model = dpf.Model(examples.static_rst)
        >>> meshed_region = model.metadata.meshed_region
        >>> nodes = meshed_region.nodes
        >>> nodes.n_nodes
        81
        
        """
        if self._nodes is None:
            self._nodes = Nodes(self)
        return self._nodes
    
    @property
    def unit(self):
        """Unit of the meshed region (same as the unit 
        of the coordinates of the meshed region)
        
        Returns
        -------
        unit : str
        """
        return self._get_unit()
    
    @unit.setter
    def unit(self, value):
        """Unit type
        
        Parameters
        ----------
        unit : str
        """
        return self._set_unit(value)

    def _get_unit(self):
        """Returns the unit type

        Returns
        -------
        unit : str
        """
        return self._stub.List(self._message).unit
    
    def _set_unit(self, unit):
        """Set the unit of the meshed_region.
        
        Parameters
        ----------
        unit: str
        """
        request = meshed_region_pb2.UpdateMeshedRegionRequest()
        request.meshed_region.CopyFrom(self._message)
        request.unit = unit
        return self._stub.UpdateRequest(request)

    def __del__(self):
        try:
            self._stub.Delete(self._message)
        except:
            pass

    def _connect(self):
        """Connect to the grpc service containing the reader"""
        return meshed_region_pb2_grpc.MeshedRegionServiceStub(self._server.channel)

    def __str__(self):
        """describe the entity
        
        Returns
        -------
        description : str
        """
        from ansys.dpf.core.core import _description
        return _description(self._message, self._server)
    
    @property
    def available_named_selections(self):
        """Returns a list of available named selections
        
        Returns
        -------
        named_selections : list str
        """
        return self._stub.List(self._message).named_selections
    
    def named_selection(self, named_selection):
        """Returns a scoping containing the list of nodes or elements
        in the named selection
        
        Parameters
        ----------
        named_selection : str 
            name of the named selection
            
        Returns
        -------
        named_selection : Scoping
        """
        if server_meet_version("2.1", self._server):
            request = meshed_region_pb2.GetScopingRequest(mesh=self._message)
            request.named_selection = named_selection
            out = self._stub.GetScoping(request)
            return scoping.Scoping(scoping=out, server=self._server)
        else:
            if hasattr(self, "_stream_provider"):
                from ansys.dpf.core.dpf_operator import Operator
                op = Operator("scoping_provider_by_ns", server = self._server)
                op.connect(1,named_selection)
                op.connect(3, self._stream_provider,0)                
                return op.get_output(0, types.scoping)
            else:
                raise Exception("getting a named selection from a meshed region is only implemented for meshed region created from a model for server version 2.0. Please update your server.")
            
    def _set_stream_provider(self, stream_provider):
        self._stream_provider = stream_provider
    # NOTE: kept only for reference as the mesh operator is being moved out of dpf
    # def write_vtk(self, filename, skin_only=True):
    #     """Return a vtk mesh"""
    #     # filename = os.path.join(tempfile.gettempdir(),
    #                             # '%s.vtk' % next(tempfile._get_candidate_names()))

    #     vtk_exp = self._model.operator("vtk_export")
    #     vtk_exp.connect(0, filename)

    #     mesh = self._model.operator("mapdl::rst::MeshProvider")
    #     mesh.connect(4, self._model.data_sources)

    #     if skin_only:
    #         skin = self._model.operator("meshed_skin_sector")
    #         skin.connect(0, mesh, 0)
    #         vtk_exp.connect(1, skin, 0)
    #     else:
    #         vtk_exp.connect(1, mesh, 0)

    #     vtk_exp.run()
    #     if not os.path.isfile(filename):
    #         raise FileNotFoundError('VTK mesh not written to disk')

    # @property
    # def skin(self):
    #     """Surface of the meshed region"""
    #     mesh = self._model.operator("mapdl::rst::MeshProvider")
    #     mesh.connect(4, self._model.data_sources)

        
    #     skin = self._model.operator("meshed_skin_sector")
    #     skin.connect(0, mesh, 0)
    #     # skin.connect(4, self)

    #     skin.get_output(0, types.meshed_region)

    #     name = None
    #     if self._name:
    #         name = 'Skin of %s' % self._name
    #     self._message = skin.get_output(0, types.meshed_region)
    #     return MeshedRegion(self._server.channel, skin, self._model, name)
    
    def _as_vtk(self, as_linear=True, include_ids=False):
        """Convert DPF mesh to a pyvista unstructured grid"""
        nodes = self.nodes.coordinates_field.data
        etypes = self.elements.element_types_field.data
        conn = self.elements.connectivities_field.data
        try:
            from ansys.dpf.core.vtk_helper import dpf_mesh_to_vtk
        except ModuleNotFoundError:
            raise ModuleNotFoundError("to use plotting capabilities, please install pyvista with :\n pip install pyvista>=0.24.0")
       
        grid = dpf_mesh_to_vtk(nodes, etypes, conn, as_linear)

        # consider adding this when scoping request is faster
        if include_ids:
            grid['node_ids'] = self.nodes.scoping.ids
            grid['element_ids'] = self.elements.scoping.ids

        return grid

    @property
    def grid(self):
        """VTK pyvista UnstructuredGrid

        Returns
        -------
        pyvista.UnstructuredGrid
            UnstructuredGrid of the mesh.

        Examples
        --------
        >>> import ansys.dpf.core as dpf
        >>> from ansys.dpf.core import examples
        >>> model = dpf.Model(examples.static_rst)
        >>> meshed_region = model.metadata.meshed_region
        >>> grid = meshed_region.grid

        Plot this grid directly

        >>> grid.plot()
        [(0.0729555495773441, 0.1029555495773441, 0.0729555495773441),
         (0.015, 0.045, 0.015),
         (0.0, 0.0, 1.0)]

        Extract the surface mesh of this grid

        >>> mesh = grid.extract_surface()
            
        """
        if self._full_grid is None:
            self._full_grid = self._as_vtk()
        return self._full_grid

    def plot(self, field_or_fields_container=None, notebook=None,
             shell_layers=None, off_screen=None, show_axes=True, **kwargs):
        """Plot the field/fields container on mesh.

        Parameters
        ----------
        field_or_fields_container
            dpf.core.Field or dpf.core.FieldsContainer

        notebook : bool, optional
            That specifies if the plotting is in the notebook (2D) or not (3D).

        shell_layers : core.shell_layers, optional
            Enum used to set the shell layers if the model to plot
            contains shell elements.

        off_screen : bool, optional
            Renders off screen when ``True``.  Useful for automated screenshots.

        show_axes : bool, optional
            Shows a vtk axes widget.  Enabled by default.

        **kwargs : optional
            Additional keyword arguments for the plotter.  See
            ``help(pyvista.plot)`` for additional keyword arguments.

        Examples
        --------
        Plot the displacement field from an example file

        >>> import ansys.dpf.core as dpf
        >>> from ansys.dpf.core import examples
        >>> model = dpf.Model(examples.static_rst)
        >>> disp = model.results.displacement()
        >>> field = disp.outputs.fields_container()[0]
        >>> model.metadata.meshed_region.plot(field)
        [(0.0729555495773441, 0.1029555495773441, 0.0729555495773441),
         (0.015, 0.045, 0.015),
         (0.0, 0.0, 1.0)]
        
        """
        pl = _DpfPlotter(self)
        if field_or_fields_container is not None:
            return pl.plot_contour(field_or_fields_container, notebook, shell_layers,
                                   off_screen, show_axes, **kwargs)

        # otherwise, simply plot self
        kwargs['notebook'] = notebook
        return pl.plot_mesh(**kwargs)
    
    def deep_copy(self,server=None):
        """Creates a deep copy of the meshed region's data on a given server.
        This can be useful to pass data from one server instance to another.
        Warning
        Only nodes scoping and coordinates and elements scoping, connectivity and types 
        are copied. The eventual property field for elemental properties and named selection
        will not be copied.
        
        Parameters
        ----------     
        server : DPFServer, optional
            Server with channel connected to the remote or local instance. When
            ``None``, attempts to use the the global server.

        Returns
        -------
        mesh_copy : MeshedRegion
        
        Examples
        --------
        >>> import ansys.dpf.core as dpf
        >>> from ansys.dpf.core import examples
        >>> model = dpf.Model(examples.static_rst)
        >>> meshed_region = model.metadata.meshed_region
        >>> other_server = dpf.start_local_server(as_global=False)
        >>> deep_copy = meshed_region.deep_copy(server=other_server)
        
        """
        node_ids = self.nodes.scoping.ids
        element_ids = self.elements.scoping.ids
        mesh = MeshedRegion(num_nodes=len(node_ids), num_elements=len(element_ids),server=server)
        with self.nodes.coordinates_field.as_local_field() as coord:
            for i,node in enumerate(mesh.nodes.add_nodes(len(node_ids))):
                node.id = node_ids[i]
                node.coordinates = coord.get_entity_data(i)
        with self.elements.connectivities_field.as_local_field() as connect :
            with self.elements.element_types_field.as_local_field() as types:
                for i,elem in enumerate(mesh.elements.add_elements(len(element_ids))):
                    elem.id = element_ids[i]
                    elem.connectivity = connect.get_entity_data(i)
                    elem.shape = element_types.shape(types.get_entity_data(i)[0])
        mesh.unit = self.unit
        return mesh
    
    
    def __send_init_request(self, num_nodes=0, num_elements=0):
        request = meshed_region_pb2.CreateRequest()
        if num_nodes:
            request.num_nodes_reserved = num_nodes
        if num_elements:
            request.num_elements_reserved = num_elements
        self._message = self._stub.Create(request)

     


"""
Operator Config
===============
"""

import functools
import grpc
from ansys import dpf
from ansys.grpc.dpf import base_pb2, operator_config_pb2, operator_config_pb2_grpc
from ansys.dpf.core.errors import protect_grpc


class Config:
    """A class used to represent an operator's configuration.
        With configurations the user can optionally choose how the operator will run.
        This is an advanced feature used for deep customization. 
        The different options can change the way loops are done, 
        it can change whether the operator needs to make check on the input or not...
    """
    def __init__(self,operator_name=None, config=None, server=None):
        if server is None:
            server = dpf.core._global_server()

        self._server = server

        self._stub = self._connect()
        
        if config:
            self._message = config
        else:
            self.__send_init_request(operator_name)
        
        if hasattr(self._message, "spec"):
            self._config_help=self._message.spec
        
        opt = self.options
        for name in opt : 
            bound_method = self.config_option_value.__get__(self, self.__class__)
            method2 = functools.partial(bound_method,
                                        config_name=name)
            setattr(self, "get_"+name+"_option", method2)
            
            bound_method = self.__set_config_option__.__get__(self, self.__class__)
            method2 = functools.partial(bound_method,
                                        config_name=name)
            setattr(self, "set_"+name+"_option", method2)
        
    def _connect(self):
        """Connect to the grpc service"""
        return operator_config_pb2_grpc.OperatorConfigServiceStub(self._server.channel)


    @protect_grpc
    def __send_init_request(self, operator_name=None):
        request = operator_config_pb2.CreateRequest()
        if operator_name:
            request.operator_name.operator_name = operator_name
        self._message = self._stub.Create(request)
    
    @property
    def options(self):
        """Returns a list of the different config options and their values"""
        tmp= self._stub.List(self._message)
        out ={}
        for opt in tmp.options:
            out[opt.option_name]=opt.value_str
        return out
    
    
    
    def __set_config_option__(self,config_value, config_name):
        """Change the value of a config option

        Parameters
        ----------
        config_name : str
            Name of the config option to change

        config_value : bool, int, float
            The value to give to this config option
        """
        request = operator_config_pb2.UpdateRequest()
        request.config.CopyFrom(self._message)
        option_request = operator_config_pb2.ConfigOption()
        option_request.option_name = config_name
        if isinstance(config_value, bool):
            option_request.bool = config_value
        elif isinstance(config_value, int):
            option_request.int = config_value
        elif isinstance(config_value, float):
            option_request.double = config_value
        else:
            raise TypeError("str, int, float are the accepted type for config options")            
        
        request.options.extend([option_request])
        self._stub.Update(request)
        
        
    def set_config_option(self, config_name, config_value):
        """Change the value of a config option

        Parameters
        ----------
        config_name : str
            Name of the config option to change

        config_value : bool, int, float
            The value to give to this config option
        """
        return self.__set_config_option__(config_value, config_name)
        
        
    def config_option_value(self, config_name):
        """Description of the given config_name and how it impacts the operator
        
        Parameters
        ----------
        config_name : str
            Name of the config option to change
            
        Returns
        ----------
        value : str
        """
        opt = self.options
        if config_name in opt:
            return opt[config_name]
        else:
            raise KeyError(f"{config_name} option doesn't exist")
            
        
    def __try_get_option__(self,config_name):
        if self._config_help:
            for option in self._config_help.config_options_spec:
                if option.name==config_name:
                    return option
        return None
    
    
    def config_option_documentation(self, config_name):
        """Description of the given config_name and how it impacts the operator
        
        Parameters
        ----------
        config_name : str
            Name of the config option to change
            
        Returns
        ----------
        documentation : str
        """
        option =self.__try_get_option__(config_name)
        if option:
            return option.document
        return ""
    
    
    def config_option_accepted_types(self, config_name):
        """Description of the given config_name and how it impacts the operator
        
        Parameters
        ----------
        config_name : str
            Name of the config option to change
            
        Returns
        ----------
        types : list str
        """
        if self._config_help:
            for option in self._config_help.config_options_spec:
                if option.name==config_name:
                    return option.type_names
        return ""
    
    def config_option_default_value(self, config_name):
        """Description of the given config_name and how it impacts the operator
        
        Parameters
        ----------
        config_name : str
            Name of the config option to change
            
        Returns
        ----------
        default value : str
        """
        if self._config_help:
            for option in self._config_help.config_options_spec:
                if option.name==config_name:
                    return option.default_value_str
        return ""
    
    
    @property
    def available_config_options(self):
        """Returns the list of available config options for this operator
        
        Returns
        ----------
        types : list str
        """
        tmp= self._stub.List(self._message)
        out =[]
        for opt in tmp.options:
            out.append(opt.option_name)
        return out

    def __str__(self):
        """describe the entity
        
        Returns
        -------
        description : str
        """
        from ansys.dpf.core.core import _description
        return _description(self._message, self._server)
    


import os
import inspect

if os.environ.get('DPF_DOCKER', '').lower() == 'true':
    # must pass a path that can be accessed by a docker image with
    # this directory mounted at the repository level for CI
    _module_path = r'/dpf/ansys/dpf/core/examples/'
else:
    _module_path = os.path.dirname(inspect.getfile(inspect.currentframe()))

# this files can be imported with from `ansys.dpf.core import examples`:
simple_bar = os.path.join(_module_path, 'ASimpleBar.rst')
static_rst = os.path.join(_module_path, 'static.rst')
complex_rst = os.path.join(_module_path, 'complex.rst')
multishells_rst = os.path.join(_module_path, 'model_with_ns.rst')
electric_therm = os.path.join(_module_path, 'rth', 'rth_electric.rth')
steady_therm = os.path.join(_module_path, 'rth', 'rth_steady.rth')
transient_therm = os.path.join(_module_path, 'rth', 'rth_transient.rth')
msup_transient = os.path.join(_module_path, 'msup_transient_plate1.rst')
simple_cyclic = os.path.join(_module_path, 'file_cyclic.rst')


"""
sweeping_phase_fc
=================
"""
from ansys.dpf.core.dpf_operator import Operator
from ansys.dpf.core.inputs import Input, _Inputs
from ansys.dpf.core.outputs import Output, _Outputs, _modify_output_spec_with_one_type
from ansys.dpf.core.operators.specification import PinSpecification, Specification

"""Operators from Ans.Dpf.Native plugin, from "math" category
"""

class sweeping_phase_fc(Operator):
    """Shift the phase of all the corresponding real and imaginary fields of a fields container for a given angle (in 2) of unit (in 4).

      available inputs:
        - fields_container (FieldsContainer)
        - angle (float)
        - unit_name (str)
        - abs_value (bool)

      available outputs:
        - fields_container (FieldsContainer)

      Examples
      --------
      >>> from ansys.dpf import core as dpf

      >>> # Instantiate operator
      >>> op = dpf.operators.math.sweeping_phase_fc()

      >>> # Make input connections
      >>> my_fields_container = dpf.FieldsContainer()
      >>> op.inputs.fields_container.connect(my_fields_container)
      >>> my_angle = float()
      >>> op.inputs.angle.connect(my_angle)
      >>> my_unit_name = str()
      >>> op.inputs.unit_name.connect(my_unit_name)
      >>> my_abs_value = bool()
      >>> op.inputs.abs_value.connect(my_abs_value)

      >>> # Instantiate operator and connect inputs in one line
      >>> op = dpf.operators.math.sweeping_phase_fc(fields_container=my_fields_container,angle=my_angle,unit_name=my_unit_name,abs_value=my_abs_value)

      >>> # Get output data
      >>> result_fields_container = op.outputs.fields_container()"""
    def __init__(self, fields_container=None, angle=None, unit_name=None, abs_value=None, config=None, server=None):
        super().__init__(name="sweeping_phase_fc", config = config, server = server)
        self._inputs = InputsSweepingPhaseFc(self)
        self._outputs = OutputsSweepingPhaseFc(self)
        if fields_container !=None:
            self.inputs.fields_container.connect(fields_container)
        if angle !=None:
            self.inputs.angle.connect(angle)
        if unit_name !=None:
            self.inputs.unit_name.connect(unit_name)
        if abs_value !=None:
            self.inputs.abs_value.connect(abs_value)

    @staticmethod
    def _spec():
        spec = Specification(description="""Shift the phase of all the corresponding real and imaginary fields of a fields container for a given angle (in 2) of unit (in 4).""",
                             map_input_pin_spec={
                                 0 : PinSpecification(name = "fields_container", type_names=["fields_container"], optional=False, document=""""""), 
                                 2 : PinSpecification(name = "angle", type_names=["double"], optional=False, document=""""""), 
                                 3 : PinSpecification(name = "unit_name", type_names=["string"], optional=False, document="""String Unit"""), 
                                 4 : PinSpecification(name = "abs_value", type_names=["bool"], optional=False, document="""""")},
                             map_output_pin_spec={
                                 0 : PinSpecification(name = "fields_container", type_names=["fields_container"], optional=False, document="""""")})
        return spec


    @staticmethod
    def default_config():
        return Operator.default_config(name = "sweeping_phase_fc")

    @property
    def inputs(self):
        """Enables to connect inputs to the operator

        Returns
        --------
        inputs : InputsSweepingPhaseFc 
        """
        return super().inputs


    @property
    def outputs(self):
        """Enables to get outputs of the operator by evaluationg it

        Returns
        --------
        outputs : OutputsSweepingPhaseFc 
        """
        return super().outputs


#internal name: sweeping_phase_fc
#scripting name: sweeping_phase_fc
class InputsSweepingPhaseFc(_Inputs):
    """Intermediate class used to connect user inputs to sweeping_phase_fc operator

      Examples
      --------
      >>> from ansys.dpf import core as dpf

      >>> op = dpf.operators.math.sweeping_phase_fc()
      >>> my_fields_container = dpf.FieldsContainer()
      >>> op.inputs.fields_container.connect(my_fields_container)
      >>> my_angle = float()
      >>> op.inputs.angle.connect(my_angle)
      >>> my_unit_name = str()
      >>> op.inputs.unit_name.connect(my_unit_name)
      >>> my_abs_value = bool()
      >>> op.inputs.abs_value.connect(my_abs_value)
    """
    def __init__(self, op: Operator):
        super().__init__(sweeping_phase_fc._spec().inputs, op)
        self._fields_container = Input(sweeping_phase_fc._spec().input_pin(0), 0, op, -1) 
        self._inputs.append(self._fields_container)
        self._angle = Input(sweeping_phase_fc._spec().input_pin(2), 2, op, -1) 
        self._inputs.append(self._angle)
        self._unit_name = Input(sweeping_phase_fc._spec().input_pin(3), 3, op, -1) 
        self._inputs.append(self._unit_name)
        self._abs_value = Input(sweeping_phase_fc._spec().input_pin(4), 4, op, -1) 
        self._inputs.append(self._abs_value)

    @property
    def fields_container(self):
        """Allows to connect fields_container input to the operator

        Parameters
        ----------
        my_fields_container : FieldsContainer, 

        Examples
        --------
        >>> from ansys.dpf import core as dpf

        >>> op = dpf.operators.math.sweeping_phase_fc()
        >>> op.inputs.fields_container.connect(my_fields_container)
        >>> #or
        >>> op.inputs.fields_container(my_fields_container)

        """
        return self._fields_container

    @property
    def angle(self):
        """Allows to connect angle input to the operator

        Parameters
        ----------
        my_angle : float, 

        Examples
        --------
        >>> from ansys.dpf import core as dpf

        >>> op = dpf.operators.math.sweeping_phase_fc()
        >>> op.inputs.angle.connect(my_angle)
        >>> #or
        >>> op.inputs.angle(my_angle)

        """
        return self._angle

    @property
    def unit_name(self):
        """Allows to connect unit_name input to the operator

        - pindoc: String Unit

        Parameters
        ----------
        my_unit_name : str, 

        Examples
        --------
        >>> from ansys.dpf import core as dpf

        >>> op = dpf.operators.math.sweeping_phase_fc()
        >>> op.inputs.unit_name.connect(my_unit_name)
        >>> #or
        >>> op.inputs.unit_name(my_unit_name)

        """
        return self._unit_name

    @property
    def abs_value(self):
        """Allows to connect abs_value input to the operator

        Parameters
        ----------
        my_abs_value : bool, 

        Examples
        --------
        >>> from ansys.dpf import core as dpf

        >>> op = dpf.operators.math.sweeping_phase_fc()
        >>> op.inputs.abs_value.connect(my_abs_value)
        >>> #or
        >>> op.inputs.abs_value(my_abs_value)

        """
        return self._abs_value

class OutputsSweepingPhaseFc(_Outputs):
    """Intermediate class used to get outputs from sweeping_phase_fc operator
      Examples
      --------
      >>> from ansys.dpf import core as dpf

      >>> op = dpf.operators.math.sweeping_phase_fc()
      >>> # Connect inputs : op.inputs. ...
      >>> result_fields_container = op.outputs.fields_container()
    """
    def __init__(self, op: Operator):
        super().__init__(sweeping_phase_fc._spec().outputs, op)
        self._fields_container = Output(sweeping_phase_fc._spec().output_pin(0), 0, op) 
        self._outputs.append(self._fields_container)

    @property
    def fields_container(self):
        """Allows to get fields_container output of the operator


        Returns
        ----------
        my_fields_container : FieldsContainer, 

        Examples
        --------
        >>> from ansys.dpf import core as dpf

        >>> op = dpf.operators.math.sweeping_phase_fc()
        >>> # Connect inputs : op.inputs. ...
        >>> result_fields_container = op.outputs.fields_container() 
        """
        return self._fields_container




"""
external_layer
==============
"""
from ansys.dpf.core.dpf_operator import Operator
from ansys.dpf.core.inputs import Input, _Inputs
from ansys.dpf.core.outputs import Output, _Outputs, _modify_output_spec_with_one_type
from ansys.dpf.core.operators.specification import PinSpecification, Specification

"""Operators from meshOperatorsCore plugin, from "mesh" category
"""

class external_layer(Operator):
    """Extracts the external layer (thick skin) of the mesh (3D elements) in a new meshed region

      available inputs:
        - mesh (MeshedRegion)

      available outputs:
        - mesh (MeshedRegion)
        - nodes_mesh_scoping (Scoping)
        - elements_mesh_scoping (Scoping)

      Examples
      --------
      >>> from ansys.dpf import core as dpf

      >>> # Instantiate operator
      >>> op = dpf.operators.mesh.external_layer()

      >>> # Make input connections
      >>> my_mesh = dpf.MeshedRegion()
      >>> op.inputs.mesh.connect(my_mesh)

      >>> # Instantiate operator and connect inputs in one line
      >>> op = dpf.operators.mesh.external_layer(mesh=my_mesh)

      >>> # Get output data
      >>> result_mesh = op.outputs.mesh()
      >>> result_nodes_mesh_scoping = op.outputs.nodes_mesh_scoping()
      >>> result_elements_mesh_scoping = op.outputs.elements_mesh_scoping()"""
    def __init__(self, mesh=None, config=None, server=None):
        super().__init__(name="meshed_external_layer_sector", config = config, server = server)
        self._inputs = InputsExternalLayer(self)
        self._outputs = OutputsExternalLayer(self)
        if mesh !=None:
            self.inputs.mesh.connect(mesh)

    @staticmethod
    def _spec():
        spec = Specification(description="""Extracts the external layer (thick skin) of the mesh (3D elements) in a new meshed region""",
                             map_input_pin_spec={
                                 0 : PinSpecification(name = "mesh", type_names=["abstract_meshed_region"], optional=False, document="""""")},
                             map_output_pin_spec={
                                 0 : PinSpecification(name = "mesh", type_names=["abstract_meshed_region"], optional=False, document=""""""), 
                                 1 : PinSpecification(name = "nodes_mesh_scoping", type_names=["scoping"], optional=False, document=""""""), 
                                 2 : PinSpecification(name = "elements_mesh_scoping", type_names=["scoping"], optional=False, document="""""")})
        return spec


    @staticmethod
    def default_config():
        return Operator.default_config(name = "meshed_external_layer_sector")

    @property
    def inputs(self):
        """Enables to connect inputs to the operator

        Returns
        --------
        inputs : InputsExternalLayer 
        """
        return super().inputs


    @property
    def outputs(self):
        """Enables to get outputs of the operator by evaluationg it

        Returns
        --------
        outputs : OutputsExternalLayer 
        """
        return super().outputs


#internal name: meshed_external_layer_sector
#scripting name: external_layer
class InputsExternalLayer(_Inputs):
    """Intermediate class used to connect user inputs to external_layer operator

      Examples
      --------
      >>> from ansys.dpf import core as dpf

      >>> op = dpf.operators.mesh.external_layer()
      >>> my_mesh = dpf.MeshedRegion()
      >>> op.inputs.mesh.connect(my_mesh)
    """
    def __init__(self, op: Operator):
        super().__init__(external_layer._spec().inputs, op)
        self._mesh = Input(external_layer._spec().input_pin(0), 0, op, -1) 
        self._inputs.append(self._mesh)

    @property
    def mesh(self):
        """Allows to connect mesh input to the operator

        Parameters
        ----------
        my_mesh : MeshedRegion, 

        Examples
        --------
        >>> from ansys.dpf import core as dpf

        >>> op = dpf.operators.mesh.external_layer()
        >>> op.inputs.mesh.connect(my_mesh)
        >>> #or
        >>> op.inputs.mesh(my_mesh)

        """
        return self._mesh

class OutputsExternalLayer(_Outputs):
    """Intermediate class used to get outputs from external_layer operator
      Examples
      --------
      >>> from ansys.dpf import core as dpf

      >>> op = dpf.operators.mesh.external_layer()
      >>> # Connect inputs : op.inputs. ...
      >>> result_mesh = op.outputs.mesh()
      >>> result_nodes_mesh_scoping = op.outputs.nodes_mesh_scoping()
      >>> result_elements_mesh_scoping = op.outputs.elements_mesh_scoping()
    """
    def __init__(self, op: Operator):
        super().__init__(external_layer._spec().outputs, op)
        self._mesh = Output(external_layer._spec().output_pin(0), 0, op) 
        self._outputs.append(self._mesh)
        self._nodes_mesh_scoping = Output(external_layer._spec().output_pin(1), 1, op) 
        self._outputs.append(self._nodes_mesh_scoping)
        self._elements_mesh_scoping = Output(external_layer._spec().output_pin(2), 2, op) 
        self._outputs.append(self._elements_mesh_scoping)

    @property
    def mesh(self):
        """Allows to get mesh output of the operator


        Returns
        ----------
        my_mesh : MeshedRegion, 

        Examples
        --------
        >>> from ansys.dpf import core as dpf

        >>> op = dpf.operators.mesh.external_layer()
        >>> # Connect inputs : op.inputs. ...
        >>> result_mesh = op.outputs.mesh() 
        """
        return self._mesh

    @property
    def nodes_mesh_scoping(self):
        """Allows to get nodes_mesh_scoping output of the operator


        Returns
        ----------
        my_nodes_mesh_scoping : Scoping, 

        Examples
        --------
        >>> from ansys.dpf import core as dpf

        >>> op = dpf.operators.mesh.external_layer()
        >>> # Connect inputs : op.inputs. ...
        >>> result_nodes_mesh_scoping = op.outputs.nodes_mesh_scoping() 
        """
        return self._nodes_mesh_scoping

    @property
    def elements_mesh_scoping(self):
        """Allows to get elements_mesh_scoping output of the operator


        Returns
        ----------
        my_elements_mesh_scoping : Scoping, 

        Examples
        --------
        >>> from ansys.dpf import core as dpf

        >>> op = dpf.operators.mesh.external_layer()
        >>> # Connect inputs : op.inputs. ...
        >>> result_elements_mesh_scoping = op.outputs.elements_mesh_scoping() 
        """
        return self._elements_mesh_scoping




"""
field_to_fc
===========
"""
from ansys.dpf.core.dpf_operator import Operator
from ansys.dpf.core.inputs import Input, _Inputs
from ansys.dpf.core.outputs import Output, _Outputs, _modify_output_spec_with_one_type
from ansys.dpf.core.operators.specification import PinSpecification, Specification

"""Operators from Ans.Dpf.Native plugin, from "utility" category
"""

class field_to_fc(Operator):
    """Create a field container containing the field in input.

      available inputs:
        - field (Field, FieldsContainer)

      available outputs:
        - fields_container (FieldsContainer)

      Examples
      --------
      >>> from ansys.dpf import core as dpf

      >>> # Instantiate operator
      >>> op = dpf.operators.utility.field_to_fc()

      >>> # Make input connections
      >>> my_field = dpf.Field()
      >>> op.inputs.field.connect(my_field)

      >>> # Instantiate operator and connect inputs in one line
      >>> op = dpf.operators.utility.field_to_fc(field=my_field)

      >>> # Get output data
      >>> result_fields_container = op.outputs.fields_container()"""
    def __init__(self, field=None, config=None, server=None):
        super().__init__(name="InjectToFieldContainer", config = config, server = server)
        self._inputs = InputsFieldToFc(self)
        self._outputs = OutputsFieldToFc(self)
        if field !=None:
            self.inputs.field.connect(field)

    @staticmethod
    def _spec():
        spec = Specification(description="""Create a field container containing the field in input.""",
                             map_input_pin_spec={
                                 0 : PinSpecification(name = "field", type_names=["field","fields_container"], optional=False, document="""if a fields container is set in input, it is pass on as output.""")},
                             map_output_pin_spec={
                                 0 : PinSpecification(name = "fields_container", type_names=["fields_container"], optional=False, document="""""")})
        return spec


    @staticmethod
    def default_config():
        return Operator.default_config(name = "InjectToFieldContainer")

    @property
    def inputs(self):
        """Enables to connect inputs to the operator

        Returns
        --------
        inputs : InputsFieldToFc 
        """
        return super().inputs


    @property
    def outputs(self):
        """Enables to get outputs of the operator by evaluationg it

        Returns
        --------
        outputs : OutputsFieldToFc 
        """
        return super().outputs


#internal name: InjectToFieldContainer
#scripting name: field_to_fc
class InputsFieldToFc(_Inputs):
    """Intermediate class used to connect user inputs to field_to_fc operator

      Examples
      --------
      >>> from ansys.dpf import core as dpf

      >>> op = dpf.operators.utility.field_to_fc()
      >>> my_field = dpf.Field()
      >>> op.inputs.field.connect(my_field)
    """
    def __init__(self, op: Operator):
        super().__init__(field_to_fc._spec().inputs, op)
        self._field = Input(field_to_fc._spec().input_pin(0), 0, op, -1) 
        self._inputs.append(self._field)

    @property
    def field(self):
        """Allows to connect field input to the operator

        - pindoc: if a fields container is set in input, it is pass on as output.

        Parameters
        ----------
        my_field : Field, FieldsContainer, 

        Examples
        --------
        >>> from ansys.dpf import core as dpf

        >>> op = dpf.operators.utility.field_to_fc()
        >>> op.inputs.field.connect(my_field)
        >>> #or
        >>> op.inputs.field(my_field)

        """
        return self._field

class OutputsFieldToFc(_Outputs):
    """Intermediate class used to get outputs from field_to_fc operator
      Examples
      --------
      >>> from ansys.dpf import core as dpf

      >>> op = dpf.operators.utility.field_to_fc()
      >>> # Connect inputs : op.inputs. ...
      >>> result_fields_container = op.outputs.fields_container()
    """
    def __init__(self, op: Operator):
        super().__init__(field_to_fc._spec().outputs, op)
        self._fields_container = Output(field_to_fc._spec().output_pin(0), 0, op) 
        self._outputs.append(self._fields_container)

    @property
    def fields_container(self):
        """Allows to get fields_container output of the operator


        Returns
        ----------
        my_fields_container : FieldsContainer, 

        Examples
        --------
        >>> from ansys.dpf import core as dpf

        >>> op = dpf.operators.utility.field_to_fc()
        >>> # Connect inputs : op.inputs. ...
        >>> result_fields_container = op.outputs.fields_container() 
        """
        return self._fields_container




"""
Dimensionnality
===============
"""

from ansys.grpc.dpf import field_definition_pb2
from ansys.dpf.core.common import natures

class Dimensionnality:
    """Class representing the dimensionnality of the field
        Read list of dim (1D vector for scalar and vector and 2D vector for matrix)
        and create a field_definition_pb2.Dimensionality message
        
    Parameters
    ----------
    server : server.DPFServer, optional
        Server with channel connected to the remote or local instance. When
        ``None``, attempts to use the the global server.

    dim_vec : list of int
        [1]: scalar
        [3]: 3d vector
        [3,3]: matrix 3 3
        
    nature : Nature
    """
    
    def __init__(self, dim_vec, nature : natures):
        self.dim = dim_vec
        self.nature = nature
        
    def _parse_dim_to_message(self):
        message = field_definition_pb2.Dimensionality()
        message.size.extend(self.dim)
        message.nature = self.nature.value
        return message
    
    def __str__(self):
        return str(self.dim) + " "+ self.nature.name
    
    
    @staticmethod
    def scalar_dim():
        """Dimensionnality instance corresponding to a scalar field

        Returns
        -------
        dimensionnality : Dimensionnality
        """
        return Dimensionnality([1], natures.scalar)
    
    @staticmethod
    def vector_dim(size):
        """Dimensionnality instance corresponding to a vector field of size "size"

        Parameters
        ----------
        size : int
            number of components by entity
        
        Returns
        -------
        dimensionnality : Dimensionnality
        """
        return Dimensionnality([size], natures.vector)
    
    @staticmethod
    def vector_3d_dim():
        """Dimensionnality instance corresponding to a 3 dimensions vector field

        Returns
        -------
        dimensionnality : Dimensionnality
        """
        return Dimensionnality([3], natures.vector)
    
    @staticmethod
    def tensor_dim():
        """Dimensionnality instance corresponding to a symmetrical 3 3 tensor field

        Returns
        -------
        dimensionnality : Dimensionnality
        """
        return Dimensionnality([3,3], natures.symmatrix)



"""
mesh_provider
=============
"""
from ansys.dpf.core.dpf_operator import Operator
from ansys.dpf.core.inputs import Input, _Inputs
from ansys.dpf.core.outputs import Output, _Outputs, _modify_output_spec_with_one_type
from ansys.dpf.core.operators.specification import PinSpecification, Specification

"""Operators from Ans.Dpf.Native plugin, from "mesh" category
"""

class mesh_provider(Operator):
    """Read a mesh from result files and cure degenerated elements

      available inputs:
        - streams_container (StreamsContainer) (optional)
        - data_sources (DataSources)
        - read_cyclic (int) (optional)

      available outputs:
        - mesh (MeshedRegion)

      Examples
      --------
      >>> from ansys.dpf import core as dpf

      >>> # Instantiate operator
      >>> op = dpf.operators.mesh.mesh_provider()

      >>> # Make input connections
      >>> my_streams_container = dpf.StreamsContainer()
      >>> op.inputs.streams_container.connect(my_streams_container)
      >>> my_data_sources = dpf.DataSources()
      >>> op.inputs.data_sources.connect(my_data_sources)
      >>> my_read_cyclic = int()
      >>> op.inputs.read_cyclic.connect(my_read_cyclic)

      >>> # Instantiate operator and connect inputs in one line
      >>> op = dpf.operators.mesh.mesh_provider(streams_container=my_streams_container,data_sources=my_data_sources)

      >>> # Get output data
      >>> result_mesh = op.outputs.mesh()"""
    def __init__(self, streams_container=None, data_sources=None, config=None, server=None):
        super().__init__(name="MeshProvider", config = config, server = server)
        self._inputs = InputsMeshProvider(self)
        self._outputs = OutputsMeshProvider(self)
        if streams_container !=None:
            self.inputs.streams_container.connect(streams_container)
        if data_sources !=None:
            self.inputs.data_sources.connect(data_sources)

    @staticmethod
    def _spec():
        spec = Specification(description="""Read a mesh from result files and cure degenerated elements""",
                             map_input_pin_spec={
                                 3 : PinSpecification(name = "streams_container", type_names=["streams_container"], optional=True, document="""result file container allowed to be kept open to cache data"""), 
                                 4 : PinSpecification(name = "data_sources", type_names=["data_sources"], optional=False, document="""result file path container, used if no streams are set"""), 
                                 14 : PinSpecification(name = "read_cyclic", type_names=["int32"], optional=True, document="""if 1 cyclic symmetry is ignored, if 2 cyclic expansion is done (default is 1)""")},
                             map_output_pin_spec={
                                 0 : PinSpecification(name = "mesh", type_names=["abstract_meshed_region"], optional=False, document="""""")})
        return spec


    @staticmethod
    def default_config():
        return Operator.default_config(name = "MeshProvider")

    @property
    def inputs(self):
        """Enables to connect inputs to the operator

        Returns
        --------
        inputs : InputsMeshProvider 
        """
        return super().inputs


    @property
    def outputs(self):
        """Enables to get outputs of the operator by evaluationg it

        Returns
        --------
        outputs : OutputsMeshProvider 
        """
        return super().outputs


#internal name: MeshProvider
#scripting name: mesh_provider
class InputsMeshProvider(_Inputs):
    """Intermediate class used to connect user inputs to mesh_provider operator

      Examples
      --------
      >>> from ansys.dpf import core as dpf

      >>> op = dpf.operators.mesh.mesh_provider()
      >>> my_streams_container = dpf.StreamsContainer()
      >>> op.inputs.streams_container.connect(my_streams_container)
      >>> my_data_sources = dpf.DataSources()
      >>> op.inputs.data_sources.connect(my_data_sources)
      >>> my_read_cyclic = int()
      >>> op.inputs.read_cyclic.connect(my_read_cyclic)
    """
    def __init__(self, op: Operator):
        super().__init__(mesh_provider._spec().inputs, op)
        self._streams_container = Input(mesh_provider._spec().input_pin(3), 3, op, -1) 
        self._inputs.append(self._streams_container)
        self._data_sources = Input(mesh_provider._spec().input_pin(4), 4, op, -1) 
        self._inputs.append(self._data_sources)
        self._read_cyclic = Input(mesh_provider._spec().input_pin(14), 14, op, -1) 
        self._inputs.append(self._read_cyclic)

    @property
    def streams_container(self):
        """Allows to connect streams_container input to the operator

        - pindoc: result file container allowed to be kept open to cache data

        Parameters
        ----------
        my_streams_container : StreamsContainer, 

        Examples
        --------
        >>> from ansys.dpf import core as dpf

        >>> op = dpf.operators.mesh.mesh_provider()
        >>> op.inputs.streams_container.connect(my_streams_container)
        >>> #or
        >>> op.inputs.streams_container(my_streams_container)

        """
        return self._streams_container

    @property
    def data_sources(self):
        """Allows to connect data_sources input to the operator

        - pindoc: result file path container, used if no streams are set

        Parameters
        ----------
        my_data_sources : DataSources, 

        Examples
        --------
        >>> from ansys.dpf import core as dpf

        >>> op = dpf.operators.mesh.mesh_provider()
        >>> op.inputs.data_sources.connect(my_data_sources)
        >>> #or
        >>> op.inputs.data_sources(my_data_sources)

        """
        return self._data_sources

    @property
    def read_cyclic(self):
        """Allows to connect read_cyclic input to the operator

        - pindoc: if 1 cyclic symmetry is ignored, if 2 cyclic expansion is done (default is 1)

        Parameters
        ----------
        my_read_cyclic : int, 

        Examples
        --------
        >>> from ansys.dpf import core as dpf

        >>> op = dpf.operators.mesh.mesh_provider()
        >>> op.inputs.read_cyclic.connect(my_read_cyclic)
        >>> #or
        >>> op.inputs.read_cyclic(my_read_cyclic)

        """
        return self._read_cyclic

class OutputsMeshProvider(_Outputs):
    """Intermediate class used to get outputs from mesh_provider operator
      Examples
      --------
      >>> from ansys.dpf import core as dpf

      >>> op = dpf.operators.mesh.mesh_provider()
      >>> # Connect inputs : op.inputs. ...
      >>> result_mesh = op.outputs.mesh()
    """
    def __init__(self, op: Operator):
        super().__init__(mesh_provider._spec().outputs, op)
        self._mesh = Output(mesh_provider._spec().output_pin(0), 0, op) 
        self._outputs.append(self._mesh)

    @property
    def mesh(self):
        """Allows to get mesh output of the operator


        Returns
        ----------
        my_mesh : MeshedRegion, 

        Examples
        --------
        >>> from ansys.dpf import core as dpf

        >>> op = dpf.operators.mesh.mesh_provider()
        >>> # Connect inputs : op.inputs. ...
        >>> result_mesh = op.outputs.mesh() 
        """
        return self._mesh




"""
Elements
========
"""

import numpy as np
from enum import Enum


from ansys import dpf
from ansys.dpf.core import scoping, field, property_field
from ansys.grpc.dpf import meshed_region_pb2
from ansys.dpf.core.errors import protect_grpc
from ansys.dpf.core.common import locations
from ansys.dpf.core import nodes
from ansys.dpf.core.common import __write_enum_doc__

class Element:
    """An element of the mesh. Encapsulates all the properties of an element: 
        its id, index, element type, element shape, connectivity...

    Created from a ``MeshedRegion``.

    Parameters
    ----------
    mesh : MeshedRegion
        ``MeshedRegion`` containing this element.

    elementid : int
        Element ID.  This is the element number corresponding to the element.

    index : int
        Index of the element.  Fortran based index of the element in
        the result.

    nodes : list[Node]
        List of DPF nodes belonging to the element.

    Examples
    --------
    Extract a single element from a meshed region

    >>> import ansys.dpf.core as dpf
    >>> from ansys.dpf.core import examples
    >>> model = dpf.Model(examples.static_rst)
    >>> elements = model.metadata.meshed_region.elements
    >>> element = elements[0]
    

    List the coordinates belonging to the first node of the element

    >>> element.nodes[0].coordinates
    [0.015, 0.045, 0.015]
    
    """

    def __init__(self, mesh, elementid, index, nodes):
        self._id = elementid
        self._index = index
        self._nodes = nodes
        self._mesh = mesh

    @property
    def node_ids(self):
        """IDs of all the nodes in this element

        Returns
        --------
        node_ids : list[int]
            IDs of all the nodes in this element

        Examples
        --------
        >>> element.node_ids
        [1, 26, 14, 12, 2, 27, 15, 13, 33, 64, 59, 30, 37, 65, 61, 34, 28, 81, 63, 58]
        
        """
        return [node.id for node in self._nodes]

    @property
    def id(self) -> int:
        """Element number"""
        return self._id

    @property
    def index(self) -> int:
        """Index of the element in the result (int starting at 0)"""
        return self._index

    @property
    def nodes(self):
        """List of Nodes
        
        Returns
        --------
        nodes : list[Node]

        Examples
        --------
        >>> first_node = element.nodes[0]
        
        """
        return self._nodes

    @property
    def n_nodes(self) -> int:
        """Number of nodes"""
        return len(self._nodes)

    def __str__(self):
        txt = 'DPF Element %d\n' % self.id
        txt += '\tIndex:      %7d\n' % self.index
        txt += '\tNodes:      %7d\n' % self.n_nodes
        txt += f'\tType:       {self.type}\n'  
        txt += '\tShape:      %7s\n' % self.shape.capitalize()
        return txt

    @property
    def type(self) -> int:
        """Element type of this element.
        
        Returns
        -------
        element_type : element_types

        Examples
        --------
        >>> element.type
        <element_types.Hex20: 1>
        
        """
        return self._get_type()

    @protect_grpc
    def _get_type(self):
        """Return the Ansys element type"""
        request = meshed_region_pb2.ElementalPropertyRequest()
        request.mesh.CopyFrom(self._mesh._message)
        request.index = self.index
        request.property = meshed_region_pb2.ELEMENT_TYPE
        return element_types(self._mesh._stub.GetElementalProperty(request).prop)

    @property
    def shape(self) -> str:
        """Element shape.

        Can be ``'shell'``, ``'solid'``, ``'beam'``, or ``'unknown_shape'``.

        Examples
        --------
        >>> element.shape
        'solid'
        
        """
        return self._get_shape()

    @protect_grpc
    def _get_shape(self):
        """Return the element shape"""
        request = meshed_region_pb2.ElementalPropertyRequest()
        request.mesh.CopyFrom(self._mesh._message)
        request.index = self.index
        request.property = meshed_region_pb2.ELEMENT_SHAPE
        prop = self._mesh._stub.GetElementalProperty(request).prop
        return meshed_region_pb2.ElementShape.Name(prop).lower()
    
    @property
    def connectivity(self):
        """Return the ordered list of node indices of the element
        
        Returns
        --------
        connectivity : list[int]
        """
        list=[]
        for node in self._nodes:
            list.append(node.index)
        return list
        

class Elements():
    """Elements belonging to a ``meshed_region``.

    Examples
    --------
    >>> import ansys.dpf.core as dpf
    >>> from ansys.dpf.core import examples
    >>> model = dpf.Model(examples.static_rst)
    >>> elements = model.metadata.meshed_region.elements
    >>> elements.n_elements
    8
     
    """

    def __init__(self, mesh):
        self._mesh = mesh
        self._mapping_id_to_index = None

    def __str__(self):
        return 'DPF Elements object with %d elements' % len(self)

    def __getitem__(self, index):
        """Returns element based on index"""
        return self.element_by_index(index)

    def __len__(self):
        return self.n_elements

    def __iter__(self):
        for i in range(len(self)):
            yield self[i]

    def element_by_id(self, id) -> Element:
        """Return an element using its element number (id).

        Parameters
        ----------
        id : int
            Element number.

        Returns
        -------
        Element
            DPF Element

        """
        return self.__get_element(elementid=id)

    def element_by_index(self, index) -> Element:
        """Return an element using its index.

        Parameters
        ----------
        index : int
            Zero-based index.

        Returns
        -------
        element : Element.

        Examples
        --------
        elements.element_by_index(0)

        Notes
        -----
        This is equivalent to ``elements[0]``

        """
        return self.__get_element(elementindex=index)  
        
    def add_elements(self, num):     
        """Add num new elements in the mesh. 
        add_elements yields num element that the user can fill
        
        Parameters
        ----------
        num : int
            number of elements to add
        
        Returns
        -------
        yield element : ElementAdder
            element to add
        
        Examples
        --------
        >>> import ansys.dpf.core as dpf
        >>> meshed_region = dpf.MeshedRegion(num_nodes=4,num_elements=3)
        >>> i=0
        >>> for node in meshed_region.nodes.add_nodes(4):
        ...     node.id = i+1
        ...     node.coordinates = [float(i), float(i), 0.0]
        ...     i=i+1
        >>> i=0
        >>> for element in meshed_region.elements.add_elements(3):
        ...     element.id=i+1
        ...     element.connectivity = [i, i+1]
        ...     element.is_beam=True #or is_solid, is_beam, is_point
        ...     i=i+1
                
        """
        request = meshed_region_pb2.AddRequest(mesh=self._mesh._message)
        for i in range(0, num):
            add = ElementAdder()
            yield add
            element_request = meshed_region_pb2.ElementRequest(id=add.id)
            element_request.connectivity.extend(add.connectivity)          
            element_request.shape =meshed_region_pb2.ElementShape.Value(add.shape.upper())
            request.elements.append(element_request)  
        self._mesh._stub.Add(request)
    
    def add_solid_element(self, id, connectivity):
        """Appends a new solid 3D element in the mesh
        
        Parameters
        ----------
        id : int
            new element's id
        
        connectivity : list of int
            list of node indices connected to the the element
        """
        self.add_element(id,"solid",connectivity)

    def add_shell_element(self, id, connectivity):
        """Appends a new shell 2D element in the mesh
        
        Parameters
        ----------
        id : int
            new element's id
        
        connectivity : list of int
            list of node indices connected to the the element
        """
        self.add_element(id, "shell",connectivity)   
        
    def add_beam_element(self, id, connectivity):
        """Appends a new beam 1D element in the mesh
        
        Parameters
        ----------
        id : int
            new element's id
        
        connectivity : list of int
            list of node indices connected to the the element
        """
        self.add_element(id, "beam",connectivity)   
        
    def add_point_element(self, id, connectivity):
        """Appends a new point element (1 node connectivity) in the mesh
        
        Parameters
        ----------
        id : int
            new element's id
        
        connectivity : list of int
            list of node indices connected to the the element
        """
        if not isinstance(connectivity, list):
            connectivity = [connectivity]
        self.add_element(id, "unknown_shape",connectivity)
    
    def add_element(self, id, shape, connectivity):
        """Appends a new element in the mesh
        
        Parameters
        ----------
        id : int
            new element's id
            
        shape : str
            shape of the element, expected are "solid", "shell", "beam" or "unknown_shape"
        
        connectivity : list of int
            list of node indices connected to the the element
        """
        request = meshed_region_pb2.AddRequest(mesh=self._mesh._message)
        element_request = meshed_region_pb2.ElementRequest(id=id)
        element_request.connectivity.extend(connectivity)
        element_request.shape = meshed_region_pb2.ElementShape.Value(shape.upper())
        request.elements.extend([element_request])
        self._mesh._stub.Add(request)
    
    @protect_grpc
    def __get_element(self, elementindex=None, elementid=None):
        """Returns the element by its id or its index

        Parameters
        ----------
        elementid : int
            id of the requested element
        elementindex : int
            index of the requested element

        Returns
        -------
        element : Element
        """
        request = meshed_region_pb2.GetRequest()
        request.mesh.CopyFrom(self._mesh._message)
        if elementindex is None:
            request.id = elementid
        else:
            request.index = elementindex

        elementOut = self._mesh._stub.GetElement(request)
        nodesOut = []
        for node in elementOut.nodes:
            nodesOut.append(nodes.Node(self._mesh, node.id, node.index, node.coordinates))
        return Element(self._mesh, elementOut.id, elementOut.index, nodesOut)

    @property
    def scoping(self) -> scoping.Scoping:
        """The Scoping of the elements.

        Examples
        --------
        >>> my_scoping = elements.scoping
        
        """
        return self._mesh._get_scoping(loc=locations.elemental)

    @property
    def element_types_field(self):
        """Element types field

        Returns
        -------
        element_types_field : core.Field
            Field of all the element types.

        Examples
        --------
        >>> field = elements.element_types_field
        >>> field.data
        array([1, 1, 1, 1, 1, 1, 1, 1])
        
        """
        request = meshed_region_pb2.ListPropertyRequest()
        request.mesh.CopyFrom(self._mesh._message)
        # request.elemental_property = meshed_region_pb2.ElementalPropertyType.ELEMENT_TYPE
        request.elemental_property = meshed_region_pb2.ELEMENT_TYPE
        fieldOut = self._mesh._stub.ListProperty(request)
        return field.Field(server=self._mesh._server, field=fieldOut)

    @property
    @protect_grpc
    def materials_field(self):
        """Materials field

        Returns
        -------
        materials_field : core.Field
            Field of all the materials ids.

        Examples
        --------
        Extract the material ids from the materials_field

        >>> elements.materials_field.data
        array([1, 1, 1, 1, 1, 1, 1, 1])
        
        """
        request = meshed_region_pb2.ListPropertyRequest()
        request.mesh.CopyFrom(self._mesh._message)
        # request.elemental_property = meshed_region_pb2.ElementalPropertyType.MATERIAL
        request.elemental_property = meshed_region_pb2.MATERIAL
        fieldOut = self._mesh._stub.ListProperty(request)
        return field.Field(server=self._mesh._server, field=fieldOut)

    @property
    def connectivities_field(self):
        """Connectivity field : field containing for each element id
        the nodes indices connected to the given element

        Returns
        -------
        connectivities_field : Field
            Field of the node indices associated to each element.

        Examples
        --------
        >>> field = elements.connectivities_field
        >>> field.get_entity_data(1)
        array([ 0, 11, 13, 25,  2,  9,  8,  3, 29, 58, 63, 32, 40, 52, 42, 37, 28,
               55, 53, 43])

        """
        return self._get_connectivities_field()

    @protect_grpc
    def _get_connectivities_field(self):
        """Return the connectivities field"""
        request = meshed_region_pb2.ListPropertyRequest()
        request.mesh.CopyFrom(self._mesh._message)
        # request.elemental_property = meshed_region_pb2.ElementalPropertyType.CONNECTIVITY
        request.elemental_property = meshed_region_pb2.CONNECTIVITY
        fieldOut = self._mesh._stub.ListProperty(request)
        return property_field.PropertyField(server=self._mesh._server, property_field=fieldOut)

    @property
    def n_elements(self) -> int:
        """Number of elements"""
        return self._mesh._stub.List(self._mesh._message).num_element

    def _build_mapping_id_to_index(self):
        """Return a mapping between ids and indices of the entity."""
        return {eid: i for i, eid in enumerate(self.scoping.ids)}

    @property
    def mapping_id_to_index(self) -> dict:
        """Mapping between the ids and indices of the entity.

        Useful for mapping scalar results from a field to this meshed region.

        Examples
        --------
        >>> meshed_region.nodes.mapping_id_to_index
        {1: 0, 2: 1, 3: 2, 4: 3}

        """
        if self._mapping_id_to_index is None:
            self._mapping_id_to_index = self._build_mapping_id_to_index()
        return self._mapping_id_to_index

    def map_scoping(self, external_scope):
        """Return the indices to map the scoping of these elements to
        the scoping of a field.

        Parameters
        ----------
        external_scope : scoping.Scoping
            Scoping to map to.

        Returns
        -------
        indices : numpy.ndarray
            List of indices to map from the external scope to the
            scoping of these elements.

        mask : numpy.ndarray
            Members of the external scope that are in the element scoping.

        Examples
        --------
        Return the indices that map a field to an elements collection.

        >>> import ansys.dpf.core as dpf
        >>> from ansys.dpf.core import examples
        >>> model = dpf.Model(examples.static_rst)
        >>> elements = model.metadata.meshed_region.elements
        >>> vol = model.results.elemental_volume()
        >>> field = vol.outputs.fields_container()[0]
        >>> ind, mask = elements.map_scoping(field.scoping)
        >>> ind
        array([0, 1, 2, 3, 4, 5, 6, 7])
        
        """
        if external_scope.location in ['Nodal', 'NodalElemental']:
            raise ValueError('Input scope location must be "Nodal"')
        arr = np.array(list(map(self.mapping_id_to_index.get, external_scope.ids)))
        mask = arr != None
        ind = arr[mask].astype(np.int)
        return ind, mask
    
    @property
    def has_shell_elements(self) -> bool:
        """Returns true if at list one element is a 2D element (shell)"""
        return self._mesh._stub.List(self._mesh._message).element_shape_info.has_shell_elements
   
    @property
    def has_solid_elements(self) -> bool:
        """Returns true if at list one element is a 3D element (solid)"""
        return self._mesh._stub.List(self._mesh._message).element_shape_info.has_solid_elements
    
    @property
    def has_beam_elements(self) -> bool:
        """Returns true if at list one element is a 1D beam element"""
        return self._mesh._stub.List(self._mesh._message).element_shape_info.has_beam_elements
    
    @property
    def has_point_elements(self) -> bool:
        """Returns true if at list one element is a point element"""
        return self._mesh._stub.List(self._mesh._message).element_shape_info.has_point_elements
    

class ElementAdder:
    """A class used to add new elements into a meshed region
    
    Attributes
    ----------
    id : int
    
    connectivity : list of int
        ordered list of node indices of the element
        
    shape : str
        "solid", "shell", "beam" or "unknown_shape"
    
    Examples
    --------
    Create a meshed region from scratch
    
    >>> import ansys.dpf.core as dpf
    >>> meshed_region = dpf.MeshedRegion(num_nodes=4,num_elements=1)
    >>> i=0
    >>> for node in meshed_region.nodes.add_nodes(4):
    ...     node.id = i+1
    ...     node.coordinates = [float(i), float(i), 0.0]
    ...     i=i+1
    >>> for element in meshed_region.elements.add_elements(1):
    ...     element.id=1
    ...     element.connectivity = range(0,4)
    ...     element.is_shell=True #or is_solid, is_beam, is_point
    
    """
    def __init__(self):
        self.id=0
        self.connectivity=[0]
        self._shape_info={"solid":True, "shell":False, "beam":False,"point":False}
    
    @property
    def is_solid(self) ->bool :
        return self._shape_info["solid"]
    
    @is_solid.setter
    def is_solid(self, value):
        for key in self._shape_info:
            self._shape_info[key]=False
        self._shape_info["solid"]=True
    
    @property
    def is_shell(self) ->bool :
        return self._shape_info["shell"]
    
    @is_shell.setter
    def is_shell(self, value):
        for key in self._shape_info:
            self._shape_info[key]=False
        self._shape_info["shell"]=True
    
    @property
    def is_beam(self) ->bool :
        return self._shape_info["beam"]
    
    @is_beam.setter
    def is_beam(self, value):
        for key in self._shape_info:
            self._shape_info[key]=False
        self._shape_info["beam"]=True
    
    @property
    def is_point(self) ->bool :
        return self._shape_info["point"]
    
    @is_point.setter
    def is_point(self, value):
        for key in self._shape_info:
            self._shape_info[key]=False
        self._shape_info["point"]=True
     
    @property
    def shape(self) ->str:
        """Gives the element shape
        
        Returns
        --------
        shape : str
            "solid", "shell", "beam" or "unknown_shape"
        """
        if self.is_solid:
            return "solid"
        elif self.is_shell:            
            return "shell"
        elif self.is_beam:    
            return "beam"        
        else:   
            return "unknown_shape"  
    
    @shape.setter
    def shape(self, value):
        """Set the element shape
        
        Parameters
        --------
        shape : str
            "solid", "shell", "beam" or "unknown_shape"
        """
        self._shape_info={"solid":False, "shell":False, "beam":False,"point":False}
        if value == "solid":
            self.is_solid=True 
        elif value =="shell":            
            self.is_shell =True
        elif value =="beam":    
            self.is_beam=True     
        else:   
            self.is_point=True
        
class element_types(Enum):
    General = -2
    All = -1
    Tet10 = 0
    Hex20 = 1
    Wedge15 = 2
    Pyramid13 = 3
    Tri6 = 4
    TriShell6 = 5
    Quad8 = 6
    QuadShell8 = 7
    Line3 = 8
    Point1 = 9
    Tet4 = 10
    Hex8 = 11
    Wedge6 = 12
    Pyramid5 = 13
    Tri3 = 14
    TriShell3 = 15
    Quad4 = 16
    QuadShell4 = 17
    Line2 = 18
    NumElementTypes = 19
    Unknown = 20
    EMagLine = 21
    EMagArc = 22
    EMagCircle = 23
    Surface3 = 24
    Surface4 = 25
    Surface6 = 26
    Surface8 = 27
    Edge2 = 28
    Edge3 = 29
    Beam3 = 30
    Beam4 = 31
    GeneralPlaceholder = 32
    Polygon = 33
    Polyhedron=34
    
    @staticmethod
    def _shapes():
        return { element_types.Tet10 : "solid",
                element_types.Hex20 : "solid",
                element_types.Wedge15 : "solid",
                element_types.Pyramid13 : "solid",
                element_types.Tri6 : "shell",
                element_types.TriShell6 :  "shell",
                element_types.Quad8 : "shell",
                element_types.QuadShell8 : "shell",
                element_types.Line3 : "beam",
                element_types.Tet4 : "solid",
                element_types.Hex8 : "solid",
                element_types.Wedge6 : "solid",
                element_types.Pyramid5 : "solid",
                element_types.Tri3 : "shell",
                element_types.TriShell3 : "shell",
                element_types.Quad4 : "shell",
                element_types.QuadShell4 : "shell",
                element_types.Line2 : "beam",
                element_types.EMagLine : "beam",
                element_types.EMagArc: "beam",
                element_types.EMagCircle : "shell",
                element_types.Surface3 : "shell",
                element_types.Surface4 : "shell",
                element_types.Surface6 : "shell",
                element_types.Surface8 : "shell",
                element_types.Edge2 : "beam",
                element_types.Edge3 : "beam",
                element_types.Beam3 : "beam",
                element_types.Beam4 : "beam",
                element_types.Polygon : "shell",
                element_types.Polyhedron : "solid"}
    
    @staticmethod
    def shape(element_type):
        if isinstance(element_type,(int, np.int32)):
            element_type = element_types(element_type)
        return element_types._shapes().get(element_type, "unknown_shape")
        

element_types.__doc__=__write_enum_doc__(element_types,"Types of elements available in a dpf's mesh.")



"""
enf_rotation_by_euler_nodes
===========================
"""
from ansys.dpf.core.dpf_operator import Operator
from ansys.dpf.core.inputs import Input, _Inputs
from ansys.dpf.core.outputs import Output, _Outputs, _modify_output_spec_with_one_type
from ansys.dpf.core.operators.specification import PinSpecification, Specification

"""Operators from mapdlOperatorsCore plugin, from "result" category
"""

class enf_rotation_by_euler_nodes(Operator):
    """read Euler angles on elements from the rst file and rotate the fields in the fieldsContainer.

      available inputs:
        - fields_container (FieldsContainer) (optional)
        - streams_container (StreamsContainer, Stream) (optional)
        - data_sources (DataSources)

      available outputs:
        - fields_container (FieldsContainer)

      Examples
      --------
      >>> from ansys.dpf import core as dpf

      >>> # Instantiate operator
      >>> op = dpf.operators.result.enf_rotation_by_euler_nodes()

      >>> # Make input connections
      >>> my_fields_container = dpf.FieldsContainer()
      >>> op.inputs.fields_container.connect(my_fields_container)
      >>> my_streams_container = dpf.StreamsContainer()
      >>> op.inputs.streams_container.connect(my_streams_container)
      >>> my_data_sources = dpf.DataSources()
      >>> op.inputs.data_sources.connect(my_data_sources)

      >>> # Instantiate operator and connect inputs in one line
      >>> op = dpf.operators.result.enf_rotation_by_euler_nodes(fields_container=my_fields_container,streams_container=my_streams_container,data_sources=my_data_sources)

      >>> # Get output data
      >>> result_fields_container = op.outputs.fields_container()"""
    def __init__(self, fields_container=None, streams_container=None, data_sources=None, config=None, server=None):
        super().__init__(name="ENF_rotation_by_euler_nodes", config = config, server = server)
        self._inputs = InputsEnfRotationByEulerNodes(self)
        self._outputs = OutputsEnfRotationByEulerNodes(self)
        if fields_container !=None:
            self.inputs.fields_container.connect(fields_container)
        if streams_container !=None:
            self.inputs.streams_container.connect(streams_container)
        if data_sources !=None:
            self.inputs.data_sources.connect(data_sources)

    @staticmethod
    def _spec():
        spec = Specification(description="""read Euler angles on elements from the rst file and rotate the fields in the fieldsContainer.""",
                             map_input_pin_spec={
                                 2 : PinSpecification(name = "fields_container", type_names=["fields_container"], optional=True, document=""""""), 
                                 3 : PinSpecification(name = "streams_container", type_names=["streams_container","stream"], optional=True, document=""""""), 
                                 4 : PinSpecification(name = "data_sources", type_names=["data_sources"], optional=False, document="""""")},
                             map_output_pin_spec={
                                 0 : PinSpecification(name = "fields_container", type_names=["fields_container"], optional=False, document="""""")})
        return spec


    @staticmethod
    def default_config():
        return Operator.default_config(name = "ENF_rotation_by_euler_nodes")

    @property
    def inputs(self):
        """Enables to connect inputs to the operator

        Returns
        --------
        inputs : InputsEnfRotationByEulerNodes 
        """
        return super().inputs


    @property
    def outputs(self):
        """Enables to get outputs of the operator by evaluationg it

        Returns
        --------
        outputs : OutputsEnfRotationByEulerNodes 
        """
        return super().outputs


#internal name: ENF_rotation_by_euler_nodes
#scripting name: enf_rotation_by_euler_nodes
class InputsEnfRotationByEulerNodes(_Inputs):
    """Intermediate class used to connect user inputs to enf_rotation_by_euler_nodes operator

      Examples
      --------
      >>> from ansys.dpf import core as dpf

      >>> op = dpf.operators.result.enf_rotation_by_euler_nodes()
      >>> my_fields_container = dpf.FieldsContainer()
      >>> op.inputs.fields_container.connect(my_fields_container)
      >>> my_streams_container = dpf.StreamsContainer()
      >>> op.inputs.streams_container.connect(my_streams_container)
      >>> my_data_sources = dpf.DataSources()
      >>> op.inputs.data_sources.connect(my_data_sources)
    """
    def __init__(self, op: Operator):
        super().__init__(enf_rotation_by_euler_nodes._spec().inputs, op)
        self._fields_container = Input(enf_rotation_by_euler_nodes._spec().input_pin(2), 2, op, -1) 
        self._inputs.append(self._fields_container)
        self._streams_container = Input(enf_rotation_by_euler_nodes._spec().input_pin(3), 3, op, -1) 
        self._inputs.append(self._streams_container)
        self._data_sources = Input(enf_rotation_by_euler_nodes._spec().input_pin(4), 4, op, -1) 
        self._inputs.append(self._data_sources)

    @property
    def fields_container(self):
        """Allows to connect fields_container input to the operator

        Parameters
        ----------
        my_fields_container : FieldsContainer, 

        Examples
        --------
        >>> from ansys.dpf import core as dpf

        >>> op = dpf.operators.result.enf_rotation_by_euler_nodes()
        >>> op.inputs.fields_container.connect(my_fields_container)
        >>> #or
        >>> op.inputs.fields_container(my_fields_container)

        """
        return self._fields_container

    @property
    def streams_container(self):
        """Allows to connect streams_container input to the operator

        Parameters
        ----------
        my_streams_container : StreamsContainer, Stream, 

        Examples
        --------
        >>> from ansys.dpf import core as dpf

        >>> op = dpf.operators.result.enf_rotation_by_euler_nodes()
        >>> op.inputs.streams_container.connect(my_streams_container)
        >>> #or
        >>> op.inputs.streams_container(my_streams_container)

        """
        return self._streams_container

    @property
    def data_sources(self):
        """Allows to connect data_sources input to the operator

        Parameters
        ----------
        my_data_sources : DataSources, 

        Examples
        --------
        >>> from ansys.dpf import core as dpf

        >>> op = dpf.operators.result.enf_rotation_by_euler_nodes()
        >>> op.inputs.data_sources.connect(my_data_sources)
        >>> #or
        >>> op.inputs.data_sources(my_data_sources)

        """
        return self._data_sources

class OutputsEnfRotationByEulerNodes(_Outputs):
    """Intermediate class used to get outputs from enf_rotation_by_euler_nodes operator
      Examples
      --------
      >>> from ansys.dpf import core as dpf

      >>> op = dpf.operators.result.enf_rotation_by_euler_nodes()
      >>> # Connect inputs : op.inputs. ...
      >>> result_fields_container = op.outputs.fields_container()
    """
    def __init__(self, op: Operator):
        super().__init__(enf_rotation_by_euler_nodes._spec().outputs, op)
        self._fields_container = Output(enf_rotation_by_euler_nodes._spec().output_pin(0), 0, op) 
        self._outputs.append(self._fields_container)

    @property
    def fields_container(self):
        """Allows to get fields_container output of the operator


        Returns
        ----------
        my_fields_container : FieldsContainer, 

        Examples
        --------
        >>> from ansys.dpf import core as dpf

        >>> op = dpf.operators.result.enf_rotation_by_euler_nodes()
        >>> # Connect inputs : op.inputs. ...
        >>> result_fields_container = op.outputs.fields_container() 
        """
        return self._fields_container




"""
Workflow
========
Interface to underlying gRPC Workflow
"""
from textwrap import wrap
import logging
import functools

from ansys import dpf
from ansys.grpc.dpf import workflow_pb2, workflow_pb2_grpc, base_pb2
from ansys.dpf.core import (dpf_operator, inputs, outputs)
from ansys.dpf.core.common import types
from ansys.dpf.core.errors import protect_grpc

LOG = logging.getLogger(__name__)
LOG.setLevel('DEBUG')


class Workflow:
    """A class used to represent a Workflow:
        a workflow is a black box containing operators and exposing only the necessary operator's
        inputs and outputs to compute a given algorithm

    Parameters
    ----------
    server : server.DPFServer, optional
        Server with channel connected to the remote or local instance. When
        ``None``, attempts to use the global server.
        
    workflow :  workflow_pb2.Workflow

    Examples
    --------
    Create a generic Workflow computing the minimum of displacement by chaining the ``'U'``
    and ``'min_max_fc'`` operators.
    
    >>> from ansys.dpf import core as dpf
    >>> disp_op = dpf.operators.result.displacement()
    >>> max_fc_op = dpf.operators.min_max.min_max_fc(disp_op)
    >>> workflow = dpf.Workflow()
    >>> workflow.add_operators([disp_op,max_fc_op])
    >>> workflow.set_input_name("data_sources", disp_op.inputs.data_sources)
    >>> workflow.set_output_name("min", max_fc_op.outputs.field_min)
    >>> workflow.set_output_name("max", max_fc_op.outputs.field_max)
    
    
    >>> from ansys.dpf.core import examples
    >>> data_src = dpf.DataSources(examples.multishells_rst)
    >>> workflow.connect("data_sources", data_src)
    >>> min = workflow.get_output("min", dpf.types.field)
    >>> max = workflow.get_output("max", dpf.types.field)
    
    """

    def __init__(self, workflow = None, server=None):
        """Initialize the workflow by connecting to a stub.
        """
        if server is None:
            server = dpf.core._global_server()

        self._server = server
        self._stub = self._connect()

        self._message = workflow
        
        if workflow is None:
            self.__send_init_request()


    @protect_grpc
    def connect(self, pin_name, inpt, pin_out=0):
        """Connect an input on the workflow using a pin name.

        Parameters
        ----------
        pin_name : str
            Name of the pin to connect. This name should be 
            exposed before with wf.set_input_name

        inpt : str, int, double, bool, list of int, list of doubles, Field, FieldsContainer, Scoping, ScopingsContainer, 
        MeshedRegion, MeshesContainer, DataSources, Operator
            Object you wish to connect.

        pin_out : int, optional
            In case of the input is an Operator, this is the output
            pin of the input Operator.  Defaults to 0.

        Examples
        --------
        Create a generic Workflow computing the minimum of displacement by chaining the ``'U'``
        and ``'min_max_fc'`` operators.
        
        >>> from ansys.dpf import core as dpf
        >>> disp_op = dpf.operators.result.displacement()
        >>> max_fc_op = dpf.operators.min_max.min_max_fc(disp_op)
        >>> workflow = dpf.Workflow()
        >>> workflow.add_operators([disp_op,max_fc_op])
        >>> workflow.set_input_name("data_sources", disp_op.inputs.data_sources)
        >>> workflow.set_output_name("min", max_fc_op.outputs.field_min)
        >>> workflow.set_output_name("max", max_fc_op.outputs.field_max)
        
        
        >>> from ansys.dpf.core import examples
        >>> data_src = dpf.DataSources(examples.multishells_rst)
        >>> workflow.connect("data_sources", data_src)
        >>> min = workflow.get_output("min", dpf.types.field)
        >>> max = workflow.get_output("max", dpf.types.field)
        
        """
        request = workflow_pb2.UpdateConnectionRequest()
        request.wf.CopyFrom(self._message)
        request.pin_name = pin_name
        dpf_operator._fillConnectionRequestMessage(request, inpt, pin_out)
        self._stub.UpdateConnection(request)
    
            
    @protect_grpc
    def get_output(self, pin_name, output_type):
        """Returns the output of the operator on the pin number.

        Parameters
        ----------
        pin_name : str
            Name of the pin to get. This name should be 
            exposed before with wf.set_output_name

        output_type : core.type enum
            The requested type of the output.
        """
        
        request = workflow_pb2.WorkflowEvaluationRequest()
        request.wf.CopyFrom(self._message)
        request.pin_name = pin_name
        
        if output_type is not None:
            dpf_operator._write_output_type_to_proto_style(output_type, request)
            out = self._stub.Get(request)
            return dpf_operator._convertOutputMessageToPythonInstance(out, output_type, self._server)
        else:
            raise ValueError("please specify an output type to get the workflow's output")
        
        
    def set_input_name(self, name, *args):
        """Name an input pin of the workflow to expose it for future connection

        Parameters
        ----------
        name : str
            Name of the pin to connect. This name should be 
            exposed before with wf.set_input_name
            
        *args : core.Operator, core.Input, int
            operator with it's input pin number or input to name 
        
        Examples
        --------
        >>> from ansys.dpf import core as dpf
        
        >>> workflow = dpf.Workflow()
        >>> workflow.add_operators([disp_op,max_fc_op])
        >>> workflow.set_input_name("data_sources", disp_op.inputs.data_sources)      
        
        >>> from ansys.dpf.core import examples
        >>> data_src = dpf.DataSources(examples.multishells_rst)
        >>> workflow.connect("data_sources", data_src)
        
        """
        request = workflow_pb2.UpdatePinNamesRequest()
        request.wf.CopyFrom(self._message)
        input_request = workflow_pb2.OperatorNaming()
        input_request.name=name
        input_request.pin = 0
        for arg in args:
            if isinstance(arg, inputs.Input):
                input_request.pin = arg._pin
                input_request.operator.CopyFrom(arg._operator._message)
            elif isinstance(arg, dpf_operator.Operator):
                input_request.operator.CopyFrom(arg._message)
            elif isinstance(arg, int):
                input_request.pin = arg
        request.inputs_naming.extend([input_request])
        self._stub.UpdatePinNames(request)
    
    
    def set_output_name(self, name, *args):
        """Name an output pin of the workflow to expose it for future connection

        Parameters
        ----------
        name : str
            Name of the pin to connect. This name should be 
            exposed before with wf.set_input_name
            
        *args : core.Operator, core.Output, int
            operator with it's outpt pin number or output to name 
        
        Examples
        --------
        >>> from ansys.dpf import core as dpf
        
        >>> workflow = dpf.Workflow()
        >>> workflow.add_operators([disp_op,max_fc_op])
        >>> workflow.set_output_name("contour", disp_op.outputs.fields_container)      
        
        >>> fc = workflow.get_output("contour", dpf.types.fields_container)
        
        """
        request = workflow_pb2.UpdatePinNamesRequest()
        request.wf.CopyFrom(self._message)
        output_request = workflow_pb2.OperatorNaming()
        output_request.name=name
        output_request.pin = 0
        for arg in args:
            if isinstance(arg, outputs.Output):
                output_request.pin = arg._pin
                output_request.operator.CopyFrom(arg._operator._message)
            elif isinstance(arg, dpf_operator.Operator):
                output_request.operator.CopyFrom(arg._message)
            elif isinstance(arg, int):
                output_request.pin = arg
        request.outputs_naming.extend([output_request])
        self._stub.UpdatePinNames(request) 
        
    
    def add_operators(self, operators):
        """Add operators to the list of operators of the workflow

        Parameters
        ----------
        operators : dpf.core.Operator, list of dpf.core.Operator
            operators to add to the list
            
        Examples
        --------
        >>> from ansys.dpf import core as dpf
        
        >>> workflow = dpf.Workflow()
        >>> disp_op = dpf.Operator("U")
        >>> max_op = dpf.Operator("min_max")
        >>> workflow.add_operator([disp_op,max_op])
        
        """
        request = workflow_pb2.AddOperatorsRequest()
        request.wf.CopyFrom(self._message)
        if isinstance(operators, list):
            request.operators.extend([op._message for op in operators])
        elif isinstance(operators, dpf_operator.Operator):
            request.operators.extend([operators._message])
        else:
            raise TypeError(f"operators to add to the workflow are expected to be of type {type(list).__name__} or {type(dpf_operator.Operator).__name__}")
        self._stub.AddOperators(request)         
            
        
    def add_operator(self, operator):
        """Add an operator to the list of operators of the workflow

        Parameters
        ----------
        operator : dpf.core.Operator
            operator to add to the list
            
        Examples
        --------
        >>> from ansys.dpf import core as dpf
        
        >>> workflow = dpf.Workflow()
        >>> disp_op = dpf.Operator("U")
        >>> workflow.add_operator(disp_op)
        
        """
        self.add_operators(operator)
        
        
    def record(self, identifier=None, transfer_ownership = True):
        """Add the workflow to DPF's internal registry with an id returned by this method.
        The workflow can be recovered by dpf.core.Workflow.get_recorded_workflow(id)
        
        Parameters
        ----------
        identifier : str, optional
            name given to the workflow
            
        transfer_ownership : bool
            if the owner ship is not transferred, the workflow is removed from the internal registry
            as soon as the workflow has been recovered by its id
            
        Examples
        --------
        >>> from ansys.dpf import core as dpf
        
        >>> workflow = dpf.Workflow()
        >>> disp_op = dpf.Operator("U")
        >>> workflow.add_operator(disp_op)
        >>> # ...
        >>> id = workflow.record()
        >>> workflow_copy = dpf.Workflow.get_recorded_workflow(id)
        
        """
        request = workflow_pb2.RecordInInternalRegistryRequest()
        request.wf.CopyFrom(self._message)
        if identifier:
            request.identifier = identifier
        request.transferOwnership = transfer_ownership
        return self._stub.RecordInInternalRegistry(request).id 
    
    
    @staticmethod
    def get_recorded_workflow(id, server=None):
        """Recover a workflow registered (with workflow.record())
        
        Parameters
        ----------
        id : int
            id given by the method "record"
        
        Returns
        ----------
        workflow : core.Workflow()
            workflow registered in dpf's registry (server side)
            
        Examples
        --------
        >>> from ansys.dpf import core as dpf
        
        >>> workflow = dpf.Workflow()
        >>> disp_op = dpf.Operator("U")
        >>> workflow.add_operator(disp_op)
        >>> # ...
        >>> id = workflow.record()
        >>> workflow_copy = dpf.Workflow.get_recorded_workflow(id)
        
        """
        request = workflow_pb2.WorkflowFromInternalRegistryRequest()
        request.registry_id = id
        wf = Workflow(server = server)
        wf._message.CopyFrom(wf._stub.GetFromInternalRegistry(request))
        return wf
    
    
    @property
    def info(self):
        """Returns a dictionary with the operator names and the exposed input and output names
        
        Returns
        ----------
        info : dictionarry str->list str
            dictionary with "operator_names", "input_names" and "output_names" key        
        """
        tmp = self._stub.List(self._message)
        out = {"operator_names":[], "input_names":[], "output_names":[]}
        for name in tmp.operator_names:
            out["operator_names"].append(name)
        for name in tmp.input_pin_names.pin_names:
            out["input_names"].append(name)
        for name in tmp.output_pin_names.pin_names:
            out["output_names"].append(name)
        return out
    
    @property
    def operator_names(self):
        """Returns a list of the names of operators added in the workflow
        
        Returns
        ----------
        names : list str
        """
        return self.info["operator_names"]
    
    @property
    def input_names(self):
        """Returns a list of the input names exposed in the workflow with set_input_name
        
        Returns
        ----------
        names : list str
        """
        return self.info["input_names"]
    
    @property
    def output_names(self):
        """Returns a list of the output names exposed in the workflow with set_output_name
        
        Returns
        ----------
        names : list str
        """
        return self.info["output_names"]
    
    
    def chain_with(self, workflow, input_output_names=None):
        """Chain 2 workflows together so that they become one workflow
        with all the operators, inputs and outputs exposed in both workflows
        
        Parameters
        ----------
        workflow : core.Workflow
            This second workflow's inputs will be chained with this workflow's outputs
            
        input_output_names : str tuple, optional
            the input name of this workflow will be chained with the output name of the second workflow
            If nothing is specified, this workflow's outputs with the same names as the second workflow's inputs will be chained
        
        Examples
        --------
        ::
            
            +-------------------------------------------------------------------------------------------------+
            |  INPUT:                                                                                         |
            |                                                                                                 |
            |input_output_names = ("output","field" )                                                          |
            |                      ____                                  ______________________                |
    	    |  "data_sources"  -> |this| ->  "stuff"        "field" -> |workflow_to_chain_with| -> "contour"  |
    	    |"time_scoping"    -> |    |             "mesh_scoping" -> |                      |               |
    	    |                     |____| ->  "output"                  |______________________|               |
            |  OUTPUT                                                                                         |
    	    |                    ____                                                                         |
    	    |"data_sources"  -> |this| ->  "stuff"                                                            |
    	    |"time_scoping" ->  |    | ->  "contour"                                                           |
    	    |"mesh_scoping" ->  |____| -> "output"                                                             |
            +-------------------------------------------------------------------------------------------------+
           
        
        """
        request = workflow_pb2.ChainRequest()
        request.wf.CopyFrom(self._message)
        request.wf_to_chain_with.CopyFrom(workflow._message)
        if input_output_names:
            request.input_to_output.output_name = input_output_names[0]
            request.input_to_output.input_name = input_output_names[1]
        self._stub.Chain(request)            
    
        
    def _connect(self):
        """Connect to the grpc service"""
        return workflow_pb2_grpc.WorkflowServiceStub(self._server.channel)

    def __del__(self):
        try:
            self._stub.Delete(self._message)
        except:
            pass
        
        
    def __str__(self):
        """describe the entity
        
        Returns
        -------
        description : str
        """
        from ansys.dpf.core.core import _description
        return _description(self._message, self._server)
    
    @protect_grpc
    def __send_init_request(self):
        request = base_pb2.Empty()
        self._message = self._stub.Create(request)



"""
time_freq_scoping_factory
=========================

Contains functions to make easy time frequency scopings creation.
"""

from ansys.dpf import core
from ansys.dpf.core.common import natures, locations
from ansys.dpf.core import errors as dpf_errors
from ansys.dpf.core import Scoping

def scoping_by_load_step(load_step, server = None):
    """Helper function to create a specific ``ansys.dpf.core.Scoping``.
    The returned scoping will describe a specific time_freq_support element for a given load_step. 

    Parameters
    ----------
    load_step : int
        Load step id of the specific time_freq scoping
    
    server : server.DPFServer, optional
        Server with channel connected to the remote or local instance. When
        ``None``, attempts to use the the global server.   
        
    Returns
    -------
    scoping : ansys.dpf.core.Scoping
        Scoping targeting one load_step."""
    scoping = Scoping(server = server, ids = [ load_step ], location = locations.time_freq_step)
    return scoping

def scoping_by_load_steps(load_steps, server = None):
    """Helper function to create a specific ``ansys.dpf.core.Scoping``.
    The returned scoping will describe a specific time_freq_support element for a given 
    list of load_steps. 

    Parameters
    ----------
    load_steps : List of int
        Load steps ids of the specific time_freq scoping
    
    server : server.DPFServer, optional
        Server with channel connected to the remote or local instance. When
        ``None``, attempts to use the the global server.   
        
    Returns
    -------
    scoping : ansys.dpf.core.Scoping
        Scoping targeting several load_steps."""
    if not isinstance(load_steps, list):
        raise dpf_errors.InvalidTypeError("list", "load_steps") 
    scoping = Scoping(server = server, ids = load_steps, location = locations.time_freq_step)
    return scoping

def scoping_by_set(cumulative_set, server = None):
    """Helper function to create a specific ``ansys.dpf.core.Scoping``.
    The returned scoping will describe a specific time_freq_support element for a given
    cumulative set index. 

    Parameters
    ----------
    cumulative_set : int
        Cumulative index of the set
    
    server : server.DPFServer, optional
        Server with channel connected to the remote or local instance. When
        ``None``, attempts to use the the global server.   
        
    Returns
    -------
    scoping : ansys.dpf.core.Scoping
        Scoping targeting one set (referenced by cumulative index)."""
    scoping = Scoping(server = server, ids = [ cumulative_set ], location = locations.time_freq)
    return scoping

def scoping_by_sets(cumulative_sets, server = None):
    """Helper function to create a specific ``ansys.dpf.core.Scoping``.
    The returned scoping will describe a specific time_freq_support element for a given 
    list of cumulative_sets. 

    Parameters
    ----------
    cumulative_sets : List of int
        Cumulative indices of the sets 
    
    server : server.DPFServer, optional
        Server with channel connected to the remote or local instance. When
        ``None``, attempts to use the the global server.   
        
    Returns
    -------
    scoping : ansys.dpf.core.Scoping
        Scoping targeting severals sets (referenced by cumulative indices)."""
    if not isinstance(cumulative_sets, list):
        raise dpf_errors.InvalidTypeError("list", "cumulative_sets") 
    scoping = Scoping(server = server, ids = cumulative_sets, location = locations.time_freq)
    return scoping

def scoping_by_step_and_substep(load_step_id, subset_id, time_freq_support, server = None):
    """Helper function to create a specific ``ansys.dpf.core.Scoping``.
    The returned scoping will describe a specific time_freq_support element for a given 
    step, substep. 

    Parameters
    ----------
    load_step_id : int
    
    subset_id : int
    
    time_freq_support : ansys.dpf.core.TimeFreqSupport()
    
    server : server.DPFServer, optional
        Server with channel connected to the remote or local instance. When
        ``None``, attempts to use the the global server.   
        
    Returns
    -------
    scoping : ansys.dpf.core.Scoping
        Scoping based on a given step/substep of a time_freq_support."""
    set_index = time_freq_support.get_cumulative_index(load_step_id - 1, subset_id - 1)
    scoping = Scoping(ids = [ set_index + 1], location = locations.time_freq)
    return scoping

def scoping_by_step_and_substep_from_model(load_step_id, subset_id, model, server = None):
    """Helper function to create a specific ``ansys.dpf.core.Scoping``.
    The returned scoping will describe a specific model's time_freq_support element for a given 
    step, substep. 

    Parameters
    ----------
    load_step_id : int
    
    subset_id : int
    
    time_freq_support : ansys.dpf.core.TimeFreqSupport()
    
    server : server.DPFServer, optional
        Server with channel connected to the remote or local instance. When
        ``None``, attempts to use the the global server.   
        
    Returns
    -------
    scoping : ansys.dpf.core.Scoping
        Scoping based on a given step/substep of a model's time_freq_support."""
    return scoping_by_step_and_substep(load_step_id, subset_id, model.metadata.time_freq_support)